# Authors: Ronan Perry, Sambit Panda, Haoyin Xu
# Adopted from: https://github.com/neurodata/honest-forests

import threading
from numbers import Integral
from warnings import catch_warnings, simplefilter

import numpy as np
from joblib import Parallel, delayed
from sklearn.base import _fit_context, clone
from sklearn.ensemble._base import _partition_estimators, _set_random_states
from sklearn.utils import compute_sample_weight, resample
from sklearn.utils._param_validation import HasMethods, Interval, RealNotInt, StrOptions
from sklearn.utils.validation import check_is_fitted

from .._lib.sklearn.ensemble._forest import ForestClassifier
from ..tree import HonestTreeClassifier
from ._extensions import ForestClassifierMixin, _generate_sample_indices


def _parallel_build_trees(
    tree,
    bootstrap,
    X,
    y,
    sample_weight,
    tree_idx,
    n_trees,
    verbose=0,
    class_weight=None,
    n_samples_bootstrap=None,
    missing_values_in_feature_mask=None,
    classes=None,
    stratify=False,
):
    """
    Private function used to fit a single tree in parallel.

    XXX:
    1. this is copied over from scikit-learn and modified to allow sampling with
    and without replacement given ``bootstrap``.

    2. Overrides the scikit-learn implementation to allow for stratification during bootstrapping
    via the `stratify` parameter.
    """
    if verbose > 1:
        print("building tree %d of %d" % (tree_idx + 1, n_trees))

    if bootstrap:
        n_samples = X.shape[0]
        if sample_weight is None:
            curr_sample_weight = np.ones((n_samples,), dtype=np.float64)
        else:
            curr_sample_weight = sample_weight.copy()

        if stratify:
            indices = resample(
                np.arange(n_samples),
                n_samples=n_samples_bootstrap,
                stratify=y,
                replace=True,
                random_state=tree.random_state,
            )
        else:
            indices = _generate_sample_indices(
                tree.random_state, n_samples, n_samples_bootstrap, bootstrap=bootstrap
            )
        sample_counts = np.bincount(indices, minlength=n_samples)
        curr_sample_weight *= sample_counts

        if class_weight == "subsample":
            with catch_warnings():
                simplefilter("ignore", DeprecationWarning)
                curr_sample_weight *= compute_sample_weight("auto", y, indices=indices)
        elif class_weight == "balanced_subsample":
            curr_sample_weight *= compute_sample_weight("balanced", y, indices=indices)

        tree._fit(
            X,
            y,
            sample_weight=curr_sample_weight,
            check_input=False,
            missing_values_in_feature_mask=missing_values_in_feature_mask,
            classes=classes,
        )
    else:
        tree._fit(
            X,
            y,
            sample_weight=sample_weight,
            check_input=False,
            missing_values_in_feature_mask=missing_values_in_feature_mask,
            classes=classes,
        )

    return tree


class HonestForestClassifier(ForestClassifier, ForestClassifierMixin):
    """
    A forest classifier with honest leaf estimates.

    Parameters
    ----------
    n_estimators : int, default=100
        The number of trees in the forest.

    criterion : {"gini", "entropy"}, default="gini"
        The function to measure the quality of a split. Supported criteria are
        "gini" for the Gini impurity and "entropy" for the information gain.
        Note: this parameter is tree-specific.

    splitter : {"best", "random"}, default="best"
        The strategy used to choose the split at each node. Supported
        strategies are "best" to choose the best split and "random" to choose
        the best random split.

    max_depth : int, default=None
        The maximum depth of the tree. If None, then nodes are expanded until
        all leaves are pure or until all leaves contain less than
        min_samples_split samples.

    min_samples_split : int or float, default=2
        The minimum number of samples required to split an internal node:

        - If int, then consider `min_samples_split` as the minimum number.
        - If float, then `min_samples_split` is a fraction and
          `ceil(min_samples_split * n_samples)` are the minimum
          number of samples for each split.

    min_samples_leaf : int or float, default=1
        The minimum number of samples required to be at a leaf node.
        A split point at any depth will only be considered if it leaves at
        least ``min_samples_leaf`` training samples in each of the left and
        right branches.  This may have the effect of smoothing the model,
        especially in regression.

        - If int, then consider `min_samples_leaf` as the minimum number.
        - If float, then `min_samples_leaf` is a fraction and
          `ceil(min_samples_leaf * n_samples)` are the minimum
          number of samples for each node.

    min_weight_fraction_leaf : float, default=0.0
        The minimum weighted fraction of the sum total of weights (of all
        the input samples) required to be at a leaf node. Samples have
        equal weight when sample_weight is not provided.

    max_features : {"sqrt", "log2", None}, int or float, default="sqrt"
        The number of features to consider when looking for the best split:

        - If int, then consider `max_features` features at each split.
        - If float, then `max_features` is a fraction and
          `round(max_features * n_features)` features are considered at each
          split.
        - If "auto", then `max_features=sqrt(n_features)`.
        - If "sqrt", then `max_features=sqrt(n_features)`.
        - If "log2", then `max_features=log2(n_features)`.
        - If None, then `max_features=n_features`.

        Note: the search for a split does not stop until at least one
        valid partition of the node samples is found, even if it requires to
        effectively inspect more than ``max_features`` features.

    max_leaf_nodes : int, default=None
        Grow trees with ``max_leaf_nodes`` in best-first fashion.
        Best nodes are defined as relative reduction in impurity.
        If None then unlimited number of leaf nodes.

    min_impurity_decrease : float, default=0.0
        A node will be split if this split induces a decrease of the impurity
        greater than or equal to this value.

        The weighted impurity decrease equation is the following::

            N_t / N * (impurity - N_t_R / N_t * right_impurity
                                - N_t_L / N_t * left_impurity)

        where ``N`` is the total number of samples, ``N_t`` is the number of
        samples at the current node, ``N_t_L`` is the number of samples in the
        left child, and ``N_t_R`` is the number of samples in the right child.

        ``N``, ``N_t``, ``N_t_R`` and ``N_t_L`` all refer to the weighted sum,
        if ``sample_weight`` is passed.

    bootstrap : bool, default=True
        Whether bootstrap samples are used when building trees. If False, the
        whole dataset is used to build each tree.

        When bootstrap is True, each tree bootstrap samples the dataset, and then
        the unique indices are split in half, where one half is used to train
        the structure of the tree and one half is used to train the leaves of the tree.
        The remaining sample indices are considered "out of bag".

    oob_score : bool, default=False
        Whether to use out-of-bag samples to estimate the generalization score.
        Only available if bootstrap=True.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`fit`, :meth:`predict`,
        :meth:`decision_path` and :meth:`apply` are all parallelized over the
        trees. ``None`` means 1 unless in a `joblib.parallel_backend`
        context. ``-1`` means using all processors. See :term:`Glossary
        <n_jobs>` for more details.

    random_state : int, RandomState instance or None, default=None
        Controls both the randomness of the bootstrapping of the samples used
        when building trees (if ``bootstrap=True``) and the sampling of the
        features to consider when looking for the best split at each node
        (if ``max_features < n_features``).
        See :term:`Glossary <random_state>` for details.

    verbose : int, default=0
        Controls the verbosity when fitting and predicting.

    warm_start : bool, default=False
        When set to ``True``, reuse the solution of the previous call to fit
        and add more estimators to the ensemble, otherwise, just fit a whole
        new forest. See :term:`the Glossary <warm_start>`.

    class_weight : {"balanced", "balanced_subsample"}, dict or list of dicts, \
            default=None
        Weights associated with classes in the form ``{class_label: weight}``.
        If not given, all classes are supposed to have weight one. For
        multi-output problems, a list of dicts can be provided in the same
        order as the columns of y.

        Note that for multioutput (including multilabel) weights should be
        defined for each class of every column in its own dict. For example,
        for four-class multilabel classification weights should be
        [{0: 1, 1: 1}, {0: 1, 1: 5}, {0: 1, 1: 1}, {0: 1, 1: 1}] instead of
        [{1:1}, {2:5}, {3:1}, {4:1}].

        The "balanced" mode uses the values of y to automatically adjust
        weights inversely proportional to class frequencies in the input data
        as ``n_samples / (n_classes * np.bincount(y))``

        The "balanced_subsample" mode is the same as "balanced" except that
        weights are computed based on the bootstrap sample for every tree
        grown.

        For multi-output, the weights of each column of y will be multiplied.

        Note that these weights will be multiplied with sample_weight (passed
        through the fit method) if sample_weight is specified.

    ccp_alpha : non-negative float, default=0.0
        Complexity parameter used for Minimal Cost-Complexity Pruning. The
        subtree with the largest cost complexity that is smaller than
        ``ccp_alpha`` will be chosen. By default, no pruning is performed. See
        :ref:`minimal_cost_complexity_pruning` for details.

    max_samples : int or float, default=None
        If bootstrap is True, the number of samples to draw from X
        to train each base tree estimator with replacement.
        If bootstrap is False, then this will subsample
        the dataset without replacement.

        - If None (default), then draw `X.shape[0]` samples.
        - If int, then draw `max_samples` samples.
        - If float, then draw `max_samples * X.shape[0]` samples.

    honest_prior : {"ignore", "uniform", "empirical"}, default="ignore"
        Method for dealing with empty leaves during evaluation of a test
        sample. If "ignore", the tree is ignored. If "uniform", the prior tree
        posterior is 1/(number of classes). If "empirical", the prior tree
        posterior is the relative class frequency in the voting subsample.
        If all trees are ignored, the empirical estimate is returned.

    honest_fraction : float, default=0.5
        Fraction of training samples used for estimates in the trees. The
        remaining samples will be used to learn the tree structure. A larger
        fraction creates shallower trees with lower variance estimates.

    honest_method : {"prune", "apply"}, default="prune"
        Method for enforcing honesty. If "prune", the tree is pruned to enforce
        honesty. If "apply", the tree is not pruned, but the leaf estimates are
        adjusted to enforce honesty.

    kernel_method : bool, default=True
        Method for normalizing ``predict_proba`` posteriors by the number of
        samples in the leaf nodes across the forest. Contrary to the average of
        posteriors, the kernel method only normalizes the probabilities once.
        By default True.

    tree_estimator : object, default=None
        Instantiated tree of type BaseDecisionTree from treeple.
        If None, then sklearn's DecisionTreeClassifier with default parameters will
        be used. Note that none of the parameters in ``tree_estimator`` need
        to be set. The parameters of the ``tree_estimator`` can be set using
        the ``tree_estimator_params`` keyword argument.

    stratify : bool, default=True
        Whether or not to stratify sample when considering structure and leaf indices.
        This will also stratify samples when bootstrap sampling is used. For more
        information, see :func:`sklearn.utils.resample`.
        By default True.

    **tree_estimator_params : dict
        Parameters to pass to the underlying base tree estimators.
        These must be parameters for ``tree_estimator``.

    Attributes
    ----------
    estimator : treeple.tree.HonestTreeClassifier
        The child estimator template used to create the collection of fitted
        sub-estimators.

    estimators_ : list of treeple.tree.HonestTreeClassifier
        The collection of fitted sub-estimators.

    classes_ : ndarray of shape (n_classes,) or a list of such arrays
        The classes labels (single output problem), or a list of arrays of
        class labels (multi-output problem).

    n_classes_ : int or list
        The number of classes (single output problem), or a list containing the
        number of classes for each output (multi-output problem).

    n_features_in_ : int
        Number of features seen during :term:`fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during :term:`fit`. Defined only when `X`
        has feature names that are all strings.

    n_outputs_ : int
        The number of outputs when ``fit`` is performed.

    feature_importances_ : ndarray of shape (n_features,)
        The impurity-based feature importances.
        The higher, the more important the feature.
        The importance of a feature is computed as the (normalized)
        total reduction of the criterion brought by that feature.  It is also
        known as the Gini importance.

        Warning: impurity-based feature importances can be misleading for
        high cardinality features (many unique values). See
        :func:`sklearn.inspection.permutation_importance` as an alternative.

    oob_score_ : float
        Score of the training dataset obtained using an out-of-bag estimate.
        This attribute exists only when ``oob_score`` is True.

    oob_decision_function_ : ndarray of shape (n_samples, n_classes) or \
            (n_samples, n_classes, n_outputs)
        Decision function computed with out-of-bag estimate on the training
        set. If n_estimators is small it might be possible that a data point
        was never left out during the bootstrap. In this case,
        `oob_decision_function_` might contain NaN. This attribute exists
        only when ``oob_score`` is True.

    honest_decision_function_ : ndarray of shape (n_samples, n_classes) or \
            (n_samples, n_classes, n_outputs)
        Decision function computed on each sample, including only the trees
        for which it was in the honest subsample. It is possible that a sample
        is never in the honest subset in which case `honest_decision_function_`
        might contain NaN.

    structure_indices_ : list of lists, shape=(n_estimators, n_structure)
        Indices of training samples used to learn the structure.

    honest_indices_ : list of lists, shape=(n_estimators, n_honest)
        Indices of training samples used to learn leaf estimates.

    oob_samples_ : list of lists, shape=(n_estimators, n_samples_bootstrap)
        The indices of training samples that are "out-of-bag". Only used
        if ``bootstrap=True``.

    Notes
    -----
    The default values for the parameters controlling the size of the trees
    (e.g. ``max_depth``, ``min_samples_leaf``, etc.) lead to fully grown and
    unpruned trees which can potentially be very large on some data sets. To
    reduce memory consumption, the complexity and size of the trees should be
    controlled by setting those parameter values.

    The features are always randomly permuted at each split. Therefore,
    the best found split may vary, even with the same training data,
    ``max_features=n_features`` and ``bootstrap=False``, if the improvement
    of the criterion is identical for several splits enumerated during the
    search of the best split. To obtain a deterministic behaviour during
    fitting, ``random_state`` has to be fixed.

    Honesty is a feature of trees that enables unbiased estimates of confidence
    intervals. The default implementation here is using double sampling to
    implement honesty. The amount of samples used for learning split nodes vs
    leaf nodes is controlled by the ``honest_fraction`` parameter. In order to
    enforce honesty, but also enable the tree to have access to all y labels,
    we set sample_weight to 0 for a random subset of samples. This results in
    inefficiency when building trees using a greedy splitter as we still sort
    over all values of X. We recommend using propensity trees if you are
    computing causal effects.

    This forest classifier is a "meta-estimator" because any tree model can
    be used in the classification process, while enabling honesty separates
    the data used for split and leaf nodes.

    References
    ----------
    .. [1] L. Breiman, "Random Forests", Machine Learning, 45(1), 5-32, 2001.

    .. [2] S. Athey, J. Tibshirani, and S. Wager. "Generalized
            Random Forests", Annals of Statistics, 2019.

    Examples
    --------
    >>> from honest_forests.estimators import HonestForestClassifier
    >>> from sklearn.datasets import make_classification
    >>> X, y = make_classification(n_samples=1000, n_features=4,
    ...                            n_informative=2, n_redundant=0,
    ...                            random_state=0, shuffle=False)
    >>> clf = HonestForestClassifier(
    >>>        max_depth=2,
    >>>        random_state=0,
    >>>        tree_estimator=ObliqueDecisionTreeClassifier())
    >>> clf.fit(X, y)
    HonestForestClassifier(...)
    >>> print(clf.predict([[0, 0, 0, 0]]))
    [1]
    """

    _parameter_constraints: dict = {
        **ForestClassifier._parameter_constraints,
        **HonestTreeClassifier._parameter_constraints,
        "class_weight": [
            StrOptions({"balanced_subsample", "balanced"}),
            dict,
            list,
            None,
        ],
    }
    _parameter_constraints.pop("max_samples")
    _parameter_constraints["max_samples"] = [
        None,
        Interval(RealNotInt, 0.0, None, closed="right"),
        Interval(Integral, 1, None, closed="left"),
    ]
    _parameter_constraints.update(
        {
            "tree_estimator": [
                HasMethods(["fit", "predict", "predict_proba", "apply"]),
                None,
            ],
            "honest_fraction": [Interval(RealNotInt, 0.0, 1.0, closed="neither")],
            "honest_prior": [StrOptions({"empirical", "uniform", "ignore"})],
            "stratify": ["boolean"],
            "tree_estimator_params": ["dict"],
        }
    )

    def __init__(
        self,
        n_estimators=100,
        criterion="gini",
        splitter="best",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_features="sqrt",
        max_leaf_nodes=None,
        min_impurity_decrease=0.0,
        bootstrap=True,
        oob_score=False,
        n_jobs=None,
        random_state=None,
        verbose=0,
        warm_start=False,
        class_weight=None,
        ccp_alpha=0.0,
        max_samples=1.6,
        honest_prior="ignore",
        honest_fraction=0.5,
        honest_method="prune",
        kernel_method=True,
        tree_estimator=None,
        stratify=True,
        **tree_estimator_params,
    ):
        super().__init__(
            estimator=HonestTreeClassifier(),
            n_estimators=n_estimators,
            estimator_params=(
                "criterion",
                "splitter",
                "max_depth",
                "min_samples_split",
                "min_samples_leaf",
                "min_weight_fraction_leaf",
                "max_features",
                "max_leaf_nodes",
                "min_impurity_decrease",
                "random_state",
                "ccp_alpha",
                "tree_estimator",
                "honest_fraction",
                "honest_prior",
                "honest_method",
                "stratify",
                "kernel_method",
            ),
            bootstrap=bootstrap,
            oob_score=oob_score,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            warm_start=warm_start,
            class_weight=class_weight,
            max_samples=max_samples,
        )
        self.criterion = criterion
        self.splitter = splitter
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease
        self.ccp_alpha = ccp_alpha
        self.honest_fraction = honest_fraction
        self.honest_prior = honest_prior
        self.honest_method = honest_method
        self.kernel_method = kernel_method
        self.tree_estimator = tree_estimator
        self.stratify = stratify
        self._tree_estimator_params = tree_estimator_params

    @_fit_context(prefer_skip_nested_validation=True)
    def fit(self, X, y, sample_weight=None, classes=None, **fit_params):
        """
        Build a forest of trees from the training set (X, y).

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples. Internally, its dtype will be converted
            to ``dtype=np.float32``. If a sparse matrix is provided, it will be
            converted into a sparse ``csc_matrix``.

        y : array-like of shape (n_samples,) or (n_samples, n_outputs)
            The target values (class labels in classification, real numbers in
            regression).

        sample_weight : array-like of shape (n_samples,), default=None
            Sample weights. If None, then samples are equally weighted. Splits
            that would create child nodes with net zero or negative weight are
            ignored while searching for a split in each node. In the case of
            classification, splits are also ignored if they would result in any
            single class carrying a negative weight in either child node.

        classes : array-like of shape (n_classes,), default=None
            List of all the classes that can possibly appear in the y vector.

        **fit_params : dict
            Parameters to pass to the underlying base tree estimators.

                Only available if `enable_metadata_routing=True`,
                which can be set by using
                ``sklearn.set_config(enable_metadata_routing=True)``.
                See :ref:`Metadata Routing User Guide <metadata_routing>` for
                more details.

        Returns
        -------
        self : object
            Fitted estimator.
        """

        super().fit(X, y, sample_weight=sample_weight, classes=classes, **fit_params)

        # Inherit attributes from the tree estimator
        self._inherit_estimator_attributes()

        # Compute honest decision function
        self.honest_decision_function_ = self._predict_proba(
            X, indices=self.honest_indices_, impute_missing=np.nan
        )

        return self

    def _construct_trees(
        self,
        X,
        y,
        sample_weight,
        random_state,
        n_samples_bootstrap,
        missing_values_in_feature_mask,
        classes,
        n_more_estimators,
    ):
        """Override construction of trees to allow stratification during bootstrapping."""
        trees = [
            self._make_estimator(append=False, random_state=random_state)
            for i in range(n_more_estimators)
        ]

        # Parallel loop: we prefer the threading backend as the Cython code
        # for fitting the trees is internally releasing the Python GIL
        # making threading more efficient than multiprocessing in
        # that case. However, for joblib 0.12+ we respect any
        # parallel_backend contexts set at a higher level,
        # since correctness does not rely on using threads.
        trees = Parallel(
            n_jobs=self.n_jobs,
            verbose=self.verbose,
            prefer="threads",
        )(
            delayed(_parallel_build_trees)(
                t,
                self.bootstrap,
                X,
                y,
                sample_weight,
                i,
                len(trees),
                verbose=self.verbose,
                class_weight=self.class_weight,
                n_samples_bootstrap=n_samples_bootstrap,
                missing_values_in_feature_mask=missing_values_in_feature_mask,
                classes=classes,
                stratify=self.stratify,
            )
            for i, t in enumerate(trees)
        )

        # Collect newly grown trees
        self.estimators_.extend(trees)

    def _make_estimator(self, append=True, random_state=None):
        """Make and configure a copy of the `estimator_` attribute.

        Warning: This method should be used to properly instantiate new
        sub-estimators.
        """
        estimator = clone(self.estimator_)
        estimator.set_params(**{p: getattr(self, p) for p in self.estimator_params})

        # XXX: This is the only change compared to scikit-learn's make_estimator
        # additionally set the tree_estimator_params
        estimator._tree_estimator_params = self._tree_estimator_params

        if random_state is not None:
            _set_random_states(estimator, random_state)

        if append:
            self.estimators_.append(estimator)

        return estimator

    def _inherit_estimator_attributes(self):
        """Initialize necessary attributes from the provided tree estimator"""
        if hasattr(self.tree_estimator, "_inheritable_fitted_attribute"):
            for attr in self.tree_estimator._inheritable_fitted_attribute:
                setattr(self, attr, getattr(self.estimators_[0], attr))

    def predict_proba(self, X):
        """
        Predict class probabilities for X.

        The predicted class probabilities of an input sample are computed as
        the mean predicted class probabilities of the trees in the forest.
        The class probability of a single tree is the fraction of samples of
        the same class in a leaf.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The input samples. Internally, its dtype will be converted to
            ``dtype=np.float32``. If a sparse matrix is provided, it will be
            converted into a sparse ``csr_matrix``.

        Returns
        -------
        p : ndarray of shape (n_samples, n_classes), or a list of such arrays
            The class probabilities of the input samples. The order of the
            classes corresponds to that in the attribute :term:`classes_`.
        """
        return self._predict_proba(X)

    def _predict_proba(self, X, indices=None, impute_missing=np.nan):
        """predict_proba helper class"""
        check_is_fitted(self)
        X = self._validate_X_predict(X)
        n_jobs, _, _ = _partition_estimators(self.n_estimators, self.n_jobs)

        # avoid storing the output of every tree estimator by summing them here
        posteriors = [
            np.zeros((X.shape[0], j), dtype=np.float64) for j in np.atleast_1d(self.n_classes_)
        ]
        if indices is None:
            indices = [None] * self.n_estimators

        lock = threading.Lock()
        Parallel(n_jobs=n_jobs, verbose=self.verbose, require="sharedmem")(
            delayed(_accumulate_prediction)(tree.predict_proba, X, posteriors, lock, idx)
            for tree, idx in zip(self.estimators_, indices)
        )

        # Normalize to unit length, due to prior weighting
        posteriors = np.array(posteriors)
        zero_mask = posteriors.sum(2) == 0
        posteriors[~zero_mask] /= posteriors[~zero_mask].sum(1, keepdims=True)

        posteriors[zero_mask] = impute_missing

        # preserve shape of multi-outputs
        if self.n_outputs_ > 1:
            posteriors = [post for post in posteriors]

        if len(posteriors) == 1:
            return posteriors[0]
        else:
            return posteriors

    @property
    def structure_indices_(self):
        """The indices used to learn the structure of the trees."""
        check_is_fitted(self)
        return [tree.structure_indices_ for tree in self.estimators_]

    @property
    def honest_indices_(self):
        """The indices used to fit the leaf nodes."""
        check_is_fitted(self)
        return [tree.honest_indices_ for tree in self.estimators_]

    @property
    def oob_samples_(self):
        """The sample indices that are out-of-bag.

        Only utilized if ``bootstrap=True``, otherwise, all samples are "in-bag".
        """
        if self.bootstrap is False and (
            self._n_samples_bootstrap is None or self._n_samples_bootstrap == self._n_samples
        ):
            raise RuntimeError(
                "Cannot extract out-of-bag samples when bootstrap is False and "
                "n_samples == n_samples_bootstrap"
            )
        check_is_fitted(self)

        oob_samples = []

        possible_indices = np.arange(self._n_samples)
        for structure_idx, honest_idx in zip(self.structure_indices_, self.honest_indices_):
            _oob_samples = np.setdiff1d(
                possible_indices, np.concatenate((structure_idx, honest_idx))
            )
            oob_samples.append(_oob_samples)
        return oob_samples

    def __sklearn_tags__(self):
        # XXX: nans should be supportable in HRF
        tags = super().__sklearn_tags__()
        # tags.classifier_tags.multi_output = False
        tags.input_tags.allow_nan = False
        return tags

    def decision_path(self, X):
        """
        Return the decision path in the forest.

        .. versionadded:: 0.18

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The input samples. Internally, its dtype will be converted to
            ``dtype=np.float32``. If a sparse matrix is provided, it will be
            converted into a sparse ``csr_matrix``.

        Returns
        -------
        indicator : sparse matrix of shape (n_samples, n_nodes)
            Return a node indicator matrix where non zero elements indicates
            that the samples goes through the nodes. The matrix is of CSR
            format.

        n_nodes_ptr : ndarray of shape (n_estimators + 1,)
            The columns from indicator[n_nodes_ptr[i]:n_nodes_ptr[i+1]]
            gives the indicator value for the i-th estimator.
        """
        return self.estimator_.decision_path(X)

    def predict_quantiles(self, X, quantiles=0.5, method="nearest"):
        """Predict class or regression value for X at given quantiles.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            Input data.
        quantiles : float, optional
            The quantiles at which to evaluate, by default 0.5 (median).
        method : str, optional
            The method to interpolate, by default 'linear'. Can be any keyword
            argument accepted by :func:`~np.quantile`.

        Returns
        -------
        y : ndarray of shape (n_samples, n_quantiles, [n_outputs])
            The predicted values. The ``n_outputs`` dimension is present only
            for multi-output regressors.
        """
        return self.estimator_.predict_quantiles(X, quantiles, method)

    def get_leaf_node_samples(self, X):
        """Get samples in each leaf node across the forest.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The data array.

        Returns
        -------
        leaf_node_samples : array-like of shape (n_samples, n_estimators)
            Samples within each leaf node.
        """
        return self.estimator_.get_leaf_node_samples(X)

    def _get_estimators_indices(self):
        # Get drawn indices along both sample and feature axes
        for tree in self.estimators_:
            if not self.bootstrap and (
                self._n_samples_bootstrap is None or (self._n_samples_bootstrap == self._n_samples)
            ):
                yield np.arange(self._n_samples, dtype=np.int32)
            else:
                # tree.random_state is actually an immutable integer seed rather
                # than a mutable RandomState instance, so it's safe to use it
                # repeatedly when calling this property.
                seed = tree.random_state

                # Operations accessing random_state must be performed identically
                # to those in `_parallel_build_trees()`
                yield _generate_sample_indices(
                    seed, self._n_samples, self._n_samples_bootstrap, self.bootstrap
                )


def _accumulate_prediction(predict, X, out, lock, indices=None):
    """
    See https://github.com/scikit-learn/scikit-learn/blob/
    95119c13af77c76e150b753485c662b7c52a41a2/sklearn/ensemble/_forest.py#L460
    This is a utility function for joblib's Parallel.
    It cannot be placed in ForestClassifier or ForestRegressor due to joblib's
    compatibility issue with pickle.
    """

    if indices is None:
        indices = np.arange(X.shape[0])
    proba = predict(X[indices], check_input=False)

    with lock:
        if len(out) == 1:
            out[0][indices] = np.nansum([out[0][indices], proba], axis=0)
        else:
            for i in range(len(out)):
                out[i][indices] = np.nansum([out[i][indices], proba[i]], axis=0)



import threading
from collections import namedtuple
from typing import Callable
from warnings import warn

import numpy as np
from joblib import Parallel, delayed
from numpy.typing import ArrayLike
from sklearn.base import clone
from sklearn.ensemble._base import _partition_estimators
from sklearn.utils.multiclass import type_of_target

from .._lib.sklearn.ensemble._forest import ForestClassifier
from ..ensemble import HonestForestClassifier
from ..tree import MultiViewDecisionTreeClassifier
from ..tree._classes import DTYPE
from .permuteforest import PermutationHonestForestClassifier
from .utils import (
    METRIC_FUNCTIONS,
    POSITIVE_METRICS,
    _compute_null_distribution_coleman,
    _compute_null_distribution_coleman_sparse,
)


def _parallel_predict_proba_oob(predict_proba, X, out, idx, test_idx, lock):
    """
    This is a utility function for joblib's Parallel.
    It can't go locally in ForestClassifier or ForestRegressor, because joblib
    complains that it cannot pickle it when placed there.
    """
    # each tree predicts proba with a list of output (n_samples, n_classes[i])
    prediction = predict_proba(X, check_input=False)

    indices = np.zeros(X.shape[0], dtype=bool)
    indices[test_idx] = True
    with lock:
        out[idx, test_idx, :] = prediction[test_idx, :]
    return prediction


def _parallel_predict_proba_oob_sparse(predict_proba, X, test_idx):
    """
    This is a utility function for joblib's Parallel.
    It can't go locally in ForestClassifier or ForestRegressor, because joblib
    complains that it cannot pickle it when placed there. Different from
    _parallel_predict_proba_oob, this function returns the predictions and
    indices for those values that are finite (not nan or inf).
    """
    # each tree predicts proba with a list of output (n_samples, n_classes[i])
    prediction = predict_proba(X[test_idx, :], check_input=False)
    good_value_mask = np.isfinite(prediction[:, 0])
    return test_idx[good_value_mask], prediction[good_value_mask]


ForestTestResult = namedtuple(
    "ForestTestResult",
    ["observe_test_stat", "permuted_stat", "observe_stat", "pvalue", "null_dist"],
)


def build_coleman_forest(
    est: HonestForestClassifier,
    perm_est: PermutationHonestForestClassifier,
    X,
    y,
    covariate_index=None,
    metric="s@98",
    n_repeats=10_000,
    verbose=False,
    seed=None,
    return_posteriors=True,
    use_sparse=False,
    **metric_kwargs,
):
    """Build a hypothesis testing forest using a two-forest approach.

    The two-forest approach stems from the Coleman et al. 2022 paper, where
    two forests are trained: one on the original dataset, and one on the
    permuted dataset. The dataset is either permuted once, or independently for
    each tree in the permuted forest. The original test statistic is computed by
    comparing the metric on both forests ``(metric_forest - metric_perm_forest)``.
    For full details, see :footcite:`coleman2022scalable`.

    Parameters
    ----------
    est : Forest
        The type of forest to use. Must be enabled with ``bootstrap=True``.
    perm_est : Forest
        The forest to use for the permuted dataset.
    X : ArrayLike of shape (n_samples, n_features)
        Data.
    y : ArrayLike of shape (n_samples, n_outputs)
        Binary target, so ``n_outputs`` should be at most 1.
    covariate_index : ArrayLike, optional of shape (n_covariates,)
        The index array of covariates to shuffle, by default None, which
        defaults to all covariates.
    metric : str, optional
        The metric to compute, by default "s@98", for sensitivity at
        98% specificity.
    n_repeats : int, optional
        Number of times to bootstrap sample the two forests to construct
        the null distribution, by default 10000. The construction of the
        null forests will be parallelized according to the ``n_jobs``
        argument of the ``est`` forest.
    verbose : bool, optional
        Verbosity, by default False.
    seed : int, optional
        Random seed, by default None.
    return_posteriors : bool, optional
        Whether or not to return the posteriors, by default True.
    use_sparse : bool, optional
        Whether or not to use a sparse for the calculation of the permutation
        statistics, by default False. Doesn't affect return values.
    **metric_kwargs : dict, optional
        Additional keyword arguments to pass to the metric function.

    Returns
    -------
    observe_stat : float
        The test statistic. To compute the test statistic, take
        ``permute_stat_`` and subtract ``observe_stat_``.
    pvalue : float
        The p-value of the test statistic.
    orig_forest_proba : ArrayLike of shape (n_estimators, n_samples, n_outputs)
        The predicted posterior probabilities for each estimator on their
        out of bag samples.
    perm_forest_proba : ArrayLike of shape (n_estimators, n_samples, n_outputs)
        The predicted posterior probabilities for each of the permuted estimators
        on their out of bag samples.
    null_dist : ArrayLike of shape (n_repeats,)
        The null statistic differences from permuted forests.

    References
    ----------
    .. footbibliography::
    """
    metric_func: Callable[[ArrayLike, ArrayLike], float] = METRIC_FUNCTIONS[metric]

    if not isinstance(est, HonestForestClassifier):
        raise RuntimeError(f"Original forest must be a HonestForestClassifier, got {type(est)}")

    # build two sets of forests
    est, orig_forest_proba = build_oob_forest(est, X, y, use_sparse=use_sparse, verbose=verbose)

    if not isinstance(perm_est, PermutationHonestForestClassifier):
        raise RuntimeError(
            f"Permutation forest must be a PermutationHonestForestClassifier, got {type(perm_est)}"
        )

    if covariate_index is None and isinstance(est.tree_estimator, MultiViewDecisionTreeClassifier):
        warn(
            "Covariate index is not defined, but a MultiViewDecisionTreeClassifier is used. "
            "If using CoMIGHT, one should define the covariate index to permute. "
            "Defaulting to use MIGHT."
        )

    perm_est, perm_forest_proba = build_oob_forest(
        perm_est,
        X,
        y,
        use_sparse=use_sparse,
        verbose=verbose,
        covariate_index=covariate_index,
    )

    # get the number of jobs
    n_jobs = est.n_jobs

    if y.ndim == 1:
        y = y.reshape(-1, 1)

    if use_sparse:
        y_pred_proba_orig_perm, observe_stat, permute_stat, metric_star, metric_star_pi = (
            _compute_null_distribution_coleman_sparse(
                y,
                orig_forest_proba,
                perm_forest_proba,
                metric,
                n_repeats=n_repeats,
                seed=seed,
                n_jobs=n_jobs,
                **metric_kwargs,
            )
        )

        # if we are returning the posteriors, then we need to replace the
        # sparse indices and values with an array. We convert the sparse data
        # to dense data, so that the function returns results in a consistent format.
        if return_posteriors:
            n_trees = y_pred_proba_orig_perm.shape[0] // 2
            n_samples = y_pred_proba_orig_perm.shape[1]

            to_coords_data = lambda x: (x.row.astype(int), x.col.astype(int), x.data)

            row, col, data = to_coords_data(y_pred_proba_orig_perm[:n_trees, :].tocoo())
            orig_forest_proba = np.full((n_trees, n_samples), np.nan, dtype=np.float64)
            orig_forest_proba[row, col] = data

            row, col, data = to_coords_data(y_pred_proba_orig_perm[n_trees:, :].tocoo())
            perm_forest_proba = np.full((n_trees, n_samples), np.nan, dtype=np.float64)
            perm_forest_proba[row, col] = data

            if y.shape[1] == 2:
                orig_forest_proba = np.column_stack((orig_forest_proba, 1 - orig_forest_proba))
                perm_forest_proba = np.column_stack((perm_forest_proba, 1 - perm_forest_proba))
    else:
        metric_star, metric_star_pi = _compute_null_distribution_coleman(
            y,
            orig_forest_proba,
            perm_forest_proba,
            metric,
            n_repeats=n_repeats,
            seed=seed,
            n_jobs=n_jobs,
            **metric_kwargs,
        )

        y_pred_proba_orig = np.nanmean(orig_forest_proba, axis=0)
        y_pred_proba_perm = np.nanmean(perm_forest_proba, axis=0)
        observe_stat = metric_func(y, y_pred_proba_orig, **metric_kwargs)
        permute_stat = metric_func(y, y_pred_proba_perm, **metric_kwargs)

    # metric^\pi - metric = observed test statistic, which under the
    # null is normally distributed around 0
    observe_test_stat = observe_stat - permute_stat

    # metric^\pi_j - metric_j, which is centered at 0
    null_dist = metric_star_pi - metric_star

    # compute pvalue
    if metric in POSITIVE_METRICS:
        pvalue = (1 + (null_dist >= observe_test_stat).sum()) / (1 + n_repeats)
    else:
        pvalue = (1 + (null_dist <= observe_test_stat).sum()) / (1 + n_repeats)

    forest_result = ForestTestResult(
        observe_test_stat, permute_stat, observe_stat, pvalue, null_dist
    )
    if return_posteriors:
        return forest_result, orig_forest_proba, perm_forest_proba, est, perm_est
    else:
        return forest_result


def build_oob_forest(
    est: ForestClassifier, X, y, use_sparse: bool = False, verbose=False, **est_kwargs
):
    """Build a hypothesis testing forest using oob samples.

    Parameters
    ----------
    est : Forest
        The type of forest to use. Must be enabled with ``bootstrap=True``.
        The forest should have either ``oob_samples_`` or ``estimators_samples_``
        property defined, which will be used to compute the out of bag samples
        per tree.
    X : ArrayLike of shape (n_samples, n_features)
        Data.
    y : ArrayLike of shape (n_samples, n_outputs)
        Binary target, so ``n_outputs`` should be at most 1.
    use_sparse : bool, optional
        Whether or not to use a sparse representation for the posteriors.
    verbose : bool, optional
        Verbosity, by default False.
    **est_kwargs : dict, optional
        Additional keyword arguments to pass to the forest estimator ``fit`` function.

    Returns
    -------
    est : Forest
        Fitted forest.
    all_proba : ArrayLike of shape (n_estimators, n_samples, n_outputs)
        The predicted posterior probabilities for each estimator on their
        out of bag samples.
    """
    assert est.bootstrap
    assert type_of_target(y) in ("binary")
    est = clone(est)

    # build forest
    est.fit(X, y.ravel(), **est_kwargs)

    # now evaluate
    X = est._validate_X_predict(X)

    # if we trained a binning tree, then we should re-bin the data
    # XXX: this is inefficient and should be improved to be in line with what
    # the Histogram Gradient Boosting Tree does, where the binning thresholds
    # are passed into the tree itself, thus allowing us to set the node feature
    # value thresholds within the tree itself.
    if hasattr(est, "max_bins") and est.max_bins is not None:
        X = est._bin_data(X, is_training_data=False).astype(DTYPE)

    # Assign chunk of trees to jobs
    n_jobs, _, _ = _partition_estimators(est.n_estimators, est.n_jobs)

    if use_sparse:
        if hasattr(est, "oob_samples_"):
            oob_samples_list = est.oob_samples_
        else:
            inbag_samples = est.estimators_samples_
            all_samples = np.arange(X.shape[0])
            oob_samples_list = [
                np.setdiff1d(all_samples, inbag_samples[i]) for i in range(len(inbag_samples))
            ]

        oob_preds_test_idxs = Parallel(n_jobs=n_jobs, verbose=verbose, require="sharedmem")(
            delayed(_parallel_predict_proba_oob_sparse)(e.predict_proba, X, test_idx)
            for e, test_idx in zip(est.estimators_, oob_samples_list)
        )
        all_proba = list(zip(*oob_preds_test_idxs))
    else:
        # avoid storing the output of every estimator by summing them here
        lock = threading.Lock()
        # accumulate the predictions across all trees
        all_proba = np.full(
            (len(est.estimators_), X.shape[0], est.n_classes_), np.nan, dtype=np.float64
        )
        if hasattr(est, "oob_samples_"):
            Parallel(n_jobs=n_jobs, verbose=verbose, require="sharedmem")(
                delayed(_parallel_predict_proba_oob)(
                    e.predict_proba, X, all_proba, idx, test_idx, lock
                )
                for idx, (e, test_idx) in enumerate(zip(est.estimators_, est.oob_samples_))
            )
        else:
            inbag_samples = est.estimators_samples_
            all_samples = np.arange(X.shape[0])
            oob_samples_list = [
                np.setdiff1d(all_samples, inbag_samples[i]) for i in range(len(inbag_samples))
            ]
            Parallel(n_jobs=n_jobs, verbose=verbose, require="sharedmem")(
                delayed(_parallel_predict_proba_oob)(
                    e.predict_proba, X, all_proba, idx, test_idx, lock
                )
                for idx, (e, test_idx) in enumerate(zip(est.estimators_, oob_samples_list))
            )

    return est, all_proba



"""
Main Author: Haoyin Xu
Corresponding Email: haoyinxu@gmail.com
"""

# import the necessary packages
import numpy as np
from joblib import Parallel, delayed
from sklearn.metrics import accuracy_score
from sklearn.utils.multiclass import _check_partial_fit_first_call

from .._lib.sklearn.ensemble._forest import (
    RandomForestClassifier,
    _generate_sample_indices,
    _get_n_samples_bootstrap,
)
from .._lib.sklearn.tree import DecisionTreeClassifier


def _partial_fit(tree, X, y, n_samples_bootstrap, classes):
    """Internal function to partially fit a tree."""
    indices = _generate_sample_indices(tree.random_state, X.shape[0], n_samples_bootstrap)
    tree.partial_fit(X[indices, :], y[indices], classes=classes)

    return tree


class StreamDecisionForest(RandomForestClassifier):
    """
    A class used to represent a naive ensemble of
    random stream decision trees.

    Parameters
    ----------
    n_estimators : int, default=100
        An integer that represents the number of stream decision trees.

    splitter : {"best", "random"}, default="best"
        The strategy used to choose the split at each node. Supported
        strategies are "best" to choose the best split and "random" to choose
        the best random split.

    max_features : {"sqrt", "log2"}, int or float, default="sqrt"
        The number of features to consider when looking for the best split:
        - If int, then consider `max_features` features at each split.
        - If float, then `max_features` is a fraction and
          `round(max_features * n_features)` features are considered at each
          split.
        - If "sqrt", then `max_features=sqrt(n_features)`.
        - If "log2", then `max_features=log2(n_features)`.
        - If None, then `max_features=n_features`.

    bootstrap : bool, default=True
        Whether bootstrap samples are used when building trees. If False, the
        whole dataset is used to build each tree.

    n_jobs : int, default=None
        The number of jobs to run in parallel.

    max_samples : int or float, default=None
        If bootstrap is True, the number of samples to draw from X
        to train each base estimator.
        - If None (default), then draw `X.shape[0]` samples.
        - If int, then draw `max_samples` samples.
        - If float, then draw `max_samples * X.shape[0]` samples. Thus,
          `max_samples` should be in the interval `(0.0, 1.0]`.

    n_swaps : int, default=1
        The number of trees to swap at each partial fitting. The actual
        swaps occur with `1/n_batches_` probability.

    Attributes
    ----------
    estimators_ : list of sklearn.tree.DecisionTreeClassifier
        An internal list that contains all
        sklearn.tree.DecisionTreeClassifier.

    classes_ : list of all unique class labels
        An internal list that stores class labels after the first call
        to `partial_fit`.

    n_batches_ : int
        The number of batches seen with `partial_fit`.
    """

    def __init__(
        self,
        n_estimators=100,
        *,
        criterion="gini",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_features="sqrt",
        max_leaf_nodes=None,
        min_impurity_decrease=0.0,
        bootstrap=True,
        oob_score=False,
        n_jobs=None,
        random_state=None,
        verbose=0,
        warm_start=False,
        class_weight=None,
        ccp_alpha=0.0,
        max_samples=None,
        max_bins=None,
        store_leaf_values=False,
        monotonic_cst=None,
        n_swaps=1,
    ):
        super().__init__(
            n_estimators=n_estimators,
            criterion=criterion,
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_features=max_features,
            max_leaf_nodes=max_leaf_nodes,
            min_impurity_decrease=min_impurity_decrease,
            ccp_alpha=ccp_alpha,
            monotonic_cst=monotonic_cst,
            bootstrap=bootstrap,
            oob_score=oob_score,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            warm_start=warm_start,
            class_weight=class_weight,
            max_samples=max_samples,
            max_bins=max_bins,
            store_leaf_values=store_leaf_values,
        )
        self.n_swaps = n_swaps

    def fit(self, X, y, sample_weight=None, classes=None):
        """
        Partially fits the forest to data X with labels y.

        Parameters
        ----------
        X : ndarray
            Input data matrix.

        y : ndarray
            Output (i.e. response data matrix).

        classes : ndarray, default=None
            List of all the classes that can possibly appear in the y vector.
            Must be provided at the first call to partial_fit, can be omitted
            in subsequent calls.

        Returns
        -------
        self : StreamDecisionForest
            The object itself.
        """
        self.n_batches_ = 1

        return super().fit(X, y, sample_weight=sample_weight, classes=classes)

    def partial_fit(self, X, y, sample_weight=None, classes=None):
        """
        Partially fits the forest to data X with labels y.

        Parameters
        ----------
        X : {array-like, sparse matrix} of shape (n_samples, n_features)
            The training input samples. Internally, it will be converted to
            ``dtype=np.float32`` and if a sparse matrix is provided
            to a sparse ``csc_matrix``.

        y : array-like of shape (n_samples,) or (n_samples, n_outputs)
            The target values (class labels) as integers or strings.

        sample_weight : array-like of shape (n_samples,), default=None
            Sample weights. If None, then samples are equally weighted. Splits
            that would create child nodes with net zero or negative weight are
            ignored while searching for a split in each node. Splits are also
            ignored if they would result in any single class carrying a
            negative weight in either child node.

        classes : ndarray, default=None
            List of all the classes that can possibly appear in the y vector.
            Must be provided at the first call to partial_fit, can be omitted
            in subsequent calls.

        Returns
        -------
        self : StreamDecisionForest
            The object itself.
        """
        self._validate_params()

        # validate input parameters
        first_call = _check_partial_fit_first_call(self, classes=classes)

        # Fit if no tree exists yet
        if first_call:
            self.fit(
                X,
                y,
                sample_weight=sample_weight,
                classes=classes,
            )
            return self
        self.n_batches_ += 1

        if self.bootstrap:
            n_samples_bootstrap = _get_n_samples_bootstrap(X.shape[0], self.max_samples)
        else:
            n_samples_bootstrap = X.shape[0]

        # Calculate probability of swaps
        swap_prob = 1 / self.n_batches_
        if self.n_swaps > 0 and self.n_batches_ > 2 and np.random.random() <= swap_prob:
            # Evaluate forest performance
            results = Parallel(n_jobs=self.n_jobs)(
                delayed(tree.predict)(X) for tree in self.estimators_
            )

            # Sort predictions by accuracy
            acc_l = []
            for idx, result in enumerate(results):
                acc_l.append([accuracy_score(result, y), idx])
            acc_l = sorted(acc_l, key=lambda x: x[0])

            # Generate new trees
            new_trees = Parallel(n_jobs=self.n_jobs)(
                delayed(_partial_fit)(
                    DecisionTreeClassifier(
                        criterion=self.criterion,
                        splitter=self.splitter,
                        max_depth=self.max_depth,
                        min_samples_split=self.min_samples_split,
                        min_samples_leaf=self.min_samples_leaf,
                        min_weight_fraction_leaf=self.min_weight_fraction_leaf,
                        max_features=self.max_features,
                        max_leaf_nodes=self.max_leaf_nodes,
                        class_weight=self.lass_weight,
                        random_state=self.random_state,
                        min_impurity_decrease=self.min_impurity_decrease,
                        monotonic_cst=self.monotonic_cst,
                        ccp_alpha=self.ccp_alpha,
                        store_leaf_values=self.store_leaf_values,
                    ),
                    X,
                    y,
                    n_samples_bootstrap=n_samples_bootstrap,
                    classes=self.classes_,
                )
                for i in range(self.n_swaps)
            )

            # Swap worst performing trees with new trees
            for i in range(self.n_swaps):
                self.estimators_[acc_l[i][1]] = new_trees[i]

        # Update existing stream decision trees
        super().partial_fit(X, y, classes=classes)

        return self



"""
===========================
Calculating p-value (MIGHT)
===========================
"""

import matplotlib.pyplot as plt
import seaborn as sns

from treeple.datasets import make_trunk_classification
from treeple.ensemble import HonestForestClassifier
from treeple.stats import PermutationHonestForestClassifier, build_coleman_forest

sns.set(color_codes=True, style="white", context="talk", font_scale=1.5)
PALETTE = sns.color_palette("Set1")
sns.set_palette(PALETTE[1:5] + PALETTE[6:], n_colors=9)
sns.set_style("white", {"axes.edgecolor": "#dddddd"})

# %%
# Independence Testing
# --------------------
#
# Given samples from ``X`` and ``Y``, the independent hypothesis and its
# alternative are stated as:
#
# .. math:: H_0 : F_{XY} = F_X F_Y
#
# .. math:: H_A : F_{XY} \neq F_X F_Y
#
# P-value is the probability of observing a statistic more extreme than the null.
# By computing the p-value using ``treeple``, we can test if :math:`H_0`
# would be rejected, which confirms that X and Y are not independent. The p-value is
# generated by comparing the observed statistic difference with permuted
# differences, using mutual information as an example.

# %%
# MI
# --
#
# Mutual Information (*MI*) measures the mutual dependence between *X* and
# *Y*. It can be calculated by the difference between the class entropy
# (``H(Y)``) and the conditional entropy (``H(Y | X)``):
#
# .. math:: I(X; Y) = H(Y) - H(Y\mid X)
#
# Under the null hypothesis :math:`H_0`, the conditional entropy ``H(Y | X)``
# is equal to the class entropy ``H(Y)``, so the *MI* becomes zero. Thus, if
# the *MI* is significantly larger than zero, we can reject the null hypothesis
# :math:`H_0`.
#
# With a binary class simulation as an example, this tutorial will show
# how to use ``treeple`` to calculate the statistic and test the
# hypothesis with data.

# %%
# Create a simulation with two gaussians
# --------------------------------------


# create a binary class simulation with two gaussians
# 500 samples for each class, class zero is standard
# gaussian, and class one has a mean at one
X, y = make_trunk_classification(
    n_samples=1000,
    n_dim=1,
    mu_0=0,
    mu_1=1,
    n_informative=1,
    seed=1,
)

fig, ax = plt.subplots(figsize=(6, 6))
fig.tight_layout()
ax.tick_params(labelsize=15)

# histogram plot the samples
ax.hist(X[:500], bins=50, alpha=0.6, color=PALETTE[1], label="negative")
ax.hist(X[500:], bins=50, alpha=0.3, color=PALETTE[0], label="positive")
ax.set_xlabel("Variable One", fontsize=15)
ax.set_ylabel("Likelihood", fontsize=15)
plt.legend(frameon=False, fontsize=15)
plt.show()

# %%
# Fit the models and calculate the p-value
# ----------------------------------------
#
# Construct two forests, one trained with original data,
# and the other trained with permuted data. The test randomly
# permutes the two forests for a specified number of times.
#
# Each pair of forest outputs a set of mutual information statistics,
# and the statistic differences are used to calculate the p-vale.


# initialize the forest with 100 trees
est = HonestForestClassifier(
    n_estimators=100,
    max_samples=1.6,
    max_features=0.3,
    bootstrap=True,
    stratify=True,
    random_state=1,
)

# initialize another forest with 100 trees
est_null = PermutationHonestForestClassifier(
    n_estimators=100,
    max_samples=1.6,
    max_features=0.3,
    bootstrap=True,
    stratify=True,
    random_state=1,
)

PERMUTE = 10000

# conduct the hypothesis test with mutual information
observed_diff, _, _, pvalue, mix_diff = build_coleman_forest(
    est, est_null, X, y, metric="mi", n_repeats=PERMUTE, return_posteriors=False, seed=1
)

fig, ax = plt.subplots(figsize=(6, 6))
fig.tight_layout()
ax.tick_params(labelsize=15)

# histogram plot the statistic differences
ax.hist(mix_diff, bins=50, alpha=0.6, color=PALETTE[1], label="null")
ax.axvline(x=observed_diff, color=PALETTE[0], linestyle="--", label="observed")
ax.set_xlabel("Mutual Information Diff", fontsize=15)
ax.set_ylabel("# of Samples", fontsize=15)
plt.legend(frameon=False, fontsize=15)
plt.show()

print("p-value is:", round(pvalue, 2))
if pvalue < 0.05:
    print("The null hypothesis is rejected.")
else:
    print("The null hypothesis is not rejected.")
# sphinx_gallery_thumbnail_number = -1



import numpy as np
import pytest
from sklearn import datasets
from sklearn.cluster import AgglomerativeClustering
from sklearn.datasets import make_blobs
from sklearn.metrics import adjusted_rand_score
from sklearn.utils.estimator_checks import parametrize_with_checks

from treeple.ensemble import UnsupervisedObliqueRandomForest, UnsupervisedRandomForest

CLUSTER_CRITERIONS = ("twomeans", "fastbic")

FOREST_CLUSTERS = {
    "UnsupervisedRandomForest": UnsupervisedRandomForest,
    "UnsupervisedObliqueRandomForest": UnsupervisedObliqueRandomForest,
}

# load iris dataset
iris = datasets.load_iris()
rng = np.random.RandomState(1)
perm = rng.permutation(iris.target.size)
iris.data = iris.data[perm]
iris.target = iris.target[perm]


@parametrize_with_checks(
    [
        UnsupervisedRandomForest(random_state=12345, n_estimators=50),
        UnsupervisedObliqueRandomForest(random_state=12345, n_estimators=50),
    ]
)
def test_sklearn_compatible_estimator(estimator, check):
    if check.func.__name__ in [
        # Cannot apply agglomerative clustering on < 2 samples
        "check_methods_subset_invariance",
        # sample weights do not necessarily imply a sample is not used in clustering
        "check_sample_weight_equivalence",
        "check_sample_weight_equivalence_on_dense_data",
        "check_sample_weight_equivalence_on_sparse_data",
        # sample order is not preserved in predict
        "check_methods_sample_order_invariance",
    ]:
        pytest.skip()
    check(estimator)


@pytest.mark.parametrize("name, forest", FOREST_CLUSTERS.items())
@pytest.mark.parametrize("criterion", CLUSTER_CRITERIONS)
def test_check_simulation(name, forest, criterion):
    n_samples = 200
    n_classes = 2

    #
    if name == "UnsupervisedRandomForest":
        n_features = 5
        if criterion == "twomeans":
            expected_score = 0.05
        elif criterion == "fastbic":
            expected_score = 0.35
    else:
        n_features = 20

        # in the forest setting, we can overfit the training dataset perfectly
        expected_score = 1.0
    X, y = make_blobs(
        n_samples=n_samples, centers=n_classes, n_features=n_features, random_state=12345
    )

    clf = forest(criterion=criterion, random_state=12345)
    clf.fit(X)
    sim_mat = clf.compute_similarity_matrix(X)

    # all ones along the diagonal
    assert np.array_equal(sim_mat.diagonal(), np.ones(n_samples))

    cluster = AgglomerativeClustering(n_clusters=n_classes).fit(sim_mat)
    predict_labels = cluster.fit_predict(sim_mat)
    score = adjusted_rand_score(y, predict_labels)

    # XXX: This should be > 0.9 according to the UReRF. However, that could be because they used
    # the oblique projections by default
    assert (
        score >= expected_score
    ), f"{name}-blobs failed with criterion {criterion} and score = {score}"


@pytest.mark.parametrize("name, forest", FOREST_CLUSTERS.items())
@pytest.mark.parametrize("criterion", CLUSTER_CRITERIONS)
def test_check_iris(name, forest, criterion):
    # Check consistency on dataset iris.
    n_classes = 3
    est = forest(criterion=criterion, random_state=12345)
    est.fit(iris.data, iris.target)
    sim_mat = est.compute_similarity_matrix(iris.data)

    if criterion == "twomeans":
        if "oblique" in name.lower():
            expected_score = 0.21
        else:
            expected_score = 0.2
    elif criterion == "fastbic":
        if "oblique" in name.lower():
            expected_score = 0.55
        else:
            expected_score = 0.3

    cluster = AgglomerativeClustering(n_clusters=n_classes).fit(sim_mat)
    predict_labels = cluster.fit_predict(sim_mat)
    score = adjusted_rand_score(iris.target, predict_labels)

    # Two-means and fastBIC criterions perform similarly here
    assert (
        score > expected_score
    ), f"{name}-iris failed with criterion {criterion} and score = {score}"



# distutils: language = c++

# Authors: Adam Li <adam2392@gmail.com>
#          Chester Huynh <chester.huynh924@gmail.com>
#          Parth Vora <pvora4@jhu.edu>
#
# License: BSD 3 clause

# See _oblique_splitter.pyx for details.

import numpy as np

from libcpp.vector cimport vector

from ..._lib.sklearn.tree._splitter cimport SplitRecord
from ..._lib.sklearn.utils._typedefs cimport float32_t, float64_t, int8_t, intp_t, uint8_t, uint32_t
from .._oblique_splitter cimport BestObliqueSplitter, ObliqueSplitRecord

# https://github.com/cython/cython/blob/master/Cython/Includes/libcpp/algorithm.pxd
# shows how to include standard library functions in Cython
# This includes the discrete_distribution C++ class, which can be used
# to generate samples from a discrete distribution with non-uniform probabilities.
# cdef extern from "<discrete_distribution>" namespace "std" nogil:
#     cdef cppclass discrete_distribution[T]
#         ctypedef T int_type
#         ctypedef G generator_type
#         discrete_distribution(T first, T last) except +
#         operator()(&G) except +

cdef class PatchSplitter(BestObliqueSplitter):
    # The PatchSplitter creates candidate feature values by sampling 2D patches from
    # an input data vector. The input data is vectorized, so `data_height` and
    # `data_width` are used to determine the vectorized indices corresponding to
    # (x,y) coordinates in the original un-vectorized data.
    cdef public intp_t ndim                       # The number of dimensions of the input data

    cdef const intp_t[:] data_dims                      # The dimensions of the input data
    cdef const intp_t[:] min_patch_dims                 # The minimum size of the patch to sample in each dimension
    cdef const intp_t[:] max_patch_dims                 # The maximum size of the patch to sample in each dimension
    cdef const uint8_t[:] dim_contiguous            # A boolean array indicating whether each dimension is contiguous

    # TODO: check if this works and is necessary for discontiguous data
    # cdef intp_t[:] stride_offsets                # The stride offsets for each dimension
    cdef bint _discontiguous

    cdef bytes boundary                               # how to sample the patch with boundary in mind
    cdef const float32_t[:, :] feature_weight               # Whether or not to normalize each column of X when adding in a patch

    cdef intp_t[::1] _index_data_buffer
    cdef intp_t[::1] _index_patch_buffer
    cdef intp_t[:] patch_sampled_size                # A buffer to store the dimensions of the sampled patch
    cdef intp_t[:] unraveled_patch_point          # A buffer to store the unraveled patch point

    # All oblique splitters (i.e. non-axis aligned splitters) require a
    # function to sample a projection matrix that is applied to the feature matrix
    # to quickly obtain the sampled projections for candidate splits.
    cdef (intp_t, intp_t) sample_top_left_seed(
        self
    ) noexcept nogil

    cdef void sample_proj_mat(
        self,
        vector[vector[float32_t]]& proj_mat_weights,
        vector[vector[intp_t]]& proj_mat_indices
    ) noexcept nogil


# cdef class UserKernelSplitter(PatchSplitter):
#     """A class to hold user-specified kernels."""
#     cdef vector[float32_t[:, ::1]] kernel_dictionary  # A list of C-contiguous 2D kernels


cdef class GaussianKernelSplitter(PatchSplitter):
    """A class to hold Gaussian kernels.

    Overrides the weights that are generated to be sampled from a Gaussian distribution.
    See: https://www.tutorialspoint.com/gaussian-filter-generation-in-cplusplus
    See: https://gist.github.com/thomasaarholt/267ec4fff40ca9dff1106490ea3b7567
    """

    cdef void sample_proj_mat(
        self,
        vector[vector[float32_t]]& proj_mat_weights,
        vector[vector[intp_t]]& proj_mat_indices
    ) noexcept nogil



import numpy as np
import scipy.linalg
import scipy.special
import scipy.stats


def simulate_helix(
    radius_a=0,
    radius_b=1,
    obs_noise_func=None,
    nature_noise_func=None,
    alpha=0.005,
    n_samples=1000,
    return_mi_lb=False,
    random_seed=None,
):
    """Simulate data from a helix.

    Parameters
    ----------
    radius_a : int, optional
        The value of the smallest radius, by default 0.0.
    radius_b : int, optional
        The value of the largest radius, by default 1.0
    obs_noise_func : Callable, optional
        By default None, which defaults to a Uniform distribution from
        (-0.005, 0.005). If passed in, then must be a callable that when
        called returns a random number denoting the noise.
    nature_noise_func : callable, optional
        By defauult None, which will add no noise. The nature noise func
        is just an independent noise term added to ``P`` before it is
        passed to the generation of the X, Y, and Z terms.
    alpha : float, optional
        The value of the noise, by default 0.005.
    n_samples : int, optional
        Number of samples to generate, by default 1000.
    return_mi_lb : bool, optional
        Whether to return the mutual information lower bound, by default False.
    random_seed : int, optional
        The random seed.

    Returns
    -------
    P : array-like of shape (n_samples,)
        The sampled P.
    X : array-like of shape (n_samples,)
        The X dimension.
    Y : array-like of shape (n_samples,)
        The X dimension.
    Z : array-like of shape (n_samples,)
        The X dimension.
    lb : float
        The mutual information lower bound.

    Notes
    -----
    Data is generated as follows: We first sample a radius that
    defines the helix, :math:`R \\approx Unif(radius_a, radius_b)`.
    Afterwards, we generate one sample as follows::

        P = 5\\pi + 3\\pi R
        X = (P + \\epsilon_1) cos(P + \\epsilon_1) / 8\\pi + N_1
        Y = (P + \\epsilon_2) sin(P + \\epsilon_2) / 8\\pi + N_2
        Z = (P + \\epsilon_3) / 8\\pi + N_3

    where :math:`N_1,N_2,N_3` are noise variables that are independently
    sampled for each sample point. And
    :math:`\\epsilon_1, \\epsilon_2, \\epsilon_3` are "nature noise" terms
    which are off by default. This process is repeated ``n_samples`` times.

    Note, that this forms the graphical model::

        R \\rightarrow P

        P \\rightarrow X
        P \\rightarrow Y
        P \\rightarrow Z

    such that P is a confounder among X, Y and Z. This implies that X, Y and Z
    are conditionally dependent on P, whereas
    """
    rng = np.random.default_rng(random_seed)

    Radii = np.zeros((n_samples,))
    P_arr = np.zeros((n_samples,))
    X_arr = np.zeros((n_samples,))
    Y_arr = np.zeros((n_samples,))
    Z_arr = np.zeros((n_samples,))

    if obs_noise_func is None:
        obs_noise_func = lambda: rng.uniform(-alpha, alpha)
    if nature_noise_func is None:
        nature_noise_func = lambda: 0.0

    for idx in range(n_samples):
        Radii[idx] = rng.uniform(radius_a, radius_b)
        P_arr[idx] = 5 * np.pi + 3 * np.pi * Radii[idx]
        X_arr[idx] = (P_arr[idx] + nature_noise_func()) * np.cos(
            P_arr[idx] + nature_noise_func()
        ) / (8 * np.pi) + obs_noise_func()
        Y_arr[idx] = (P_arr[idx] + nature_noise_func()) * np.sin(
            P_arr[idx] + nature_noise_func()
        ) / (8 * np.pi) + obs_noise_func()
        Z_arr[idx] = (P_arr[idx] + nature_noise_func()) / (8 * np.pi) + obs_noise_func()

    if return_mi_lb:
        lb = alpha / 2 - np.log(alpha)
        return P_arr, X_arr, Y_arr, Z_arr, lb

    return P_arr, X_arr, Y_arr, Z_arr


def simulate_sphere(
    radius=1, noise_func=None, alpha=0.005, n_samples=1000, return_mi_lb=False, random_seed=None
):
    """Simulate samples generated on a sphere.

    Parameters
    ----------
    radius : int, optional
        The radius of the sphere, by default 1.
    noise_func : callable, optional
        The noise function to call to add to samples, by default None,
        which defaults to sampling from the uniform distribution [-alpha, alpha].
    alpha : float, optional
        The value of the noise, by default 0.005.
    n_samples : int, optional
        Number of samples to generate, by default 1000.
    return_mi_lb : bool, optional
        Whether to return the mutual information lower bound, by default False.
    random_seed : int, optional
        Random seed, by default None.

    Returns
    -------
    latitude : float
        Latitude.
    longitude : float
        Longitude.
    Y1 : array-like of shape (n_samples,)
        The X coordinate.
    Y2 : array-like of shape (n_samples,)
        The Y coordinate.
    Y3 : array-like of shape (n_samples,)
        The Z coordinate.
    lb : float
        The mutual information lower bound.
    """
    rng = np.random.default_rng(random_seed)
    if noise_func is None:
        noise_func = lambda: rng.uniform(-alpha, alpha)

    latitude = np.zeros((n_samples,))
    longitude = np.zeros((n_samples,))
    Y1 = np.zeros((n_samples,))
    Y2 = np.zeros((n_samples,))
    Y3 = np.zeros((n_samples,))

    # sample latitude and longitude
    for idx in range(n_samples):
        # sample latitude and longitude
        latitude[idx] = rng.uniform(0, 2 * np.pi)
        longitude[idx] = np.arccos(1 - 2 * rng.uniform(0, 1))

        Y1[idx] = np.sin(longitude[idx]) * np.cos(latitude[idx]) * radius + noise_func()
        Y2[idx] = np.sin(longitude[idx]) * np.sin(latitude[idx]) * radius + noise_func()
        Y3[idx] = np.cos(longitude[idx]) * radius + noise_func()

    if return_mi_lb:
        lb = alpha / 2 - np.log(alpha)
        return latitude, longitude, Y1, Y2, Y3, lb

    return latitude, longitude, Y1, Y2, Y3


def simulate_multivariate_gaussian(mean=None, cov=None, d=2, n_samples=1000, seed=1234):
    """Multivariate gaussian simulation for testing entropy and MI estimators.

    Simulates samples from a "known" multivariate gaussian distribution
    and then passes those samples, along with the true analytical MI/CMI.

    Parameters
    ----------
    mean : array-like of shape (n_features,)
        The optional mean array. If None (default), a random standard normal vector is drawn.
    cov : array-like of shape (n_features, n_features)
        The covariance array. If None (default), a random standard normal 2D array is drawn.
        It is then converted to a PD matrix.
    d : int
        The dimensionality of the multivariate gaussian. By default 2.
    n_samples : int
        The number of samples to generate. By default 1000.
    seed : int
        The random seed to feed to :func:`numpy.random.default_rng`.

    Returns
    -------
    data : array-like of shape (n_samples, n_features)
        The generated data from the distribution.
    mean : array-like of shape (n_features,)
        The mean vector of the distribution.
    cov : array-like of shape (n_features, n_features)
        The covariance matrix of the distribution.
    """
    rng = np.random.default_rng(seed)

    if mean is None:
        mean = rng.normal(size=(d,))
    if cov is None:
        # generate random covariance matrix and enure it is symmetric and positive-definite
        cov = rng.normal(size=(d, d))
        cov = 0.5 * (cov.dot(cov.T))
    if not np.all(np.linalg.eigvals(cov) > 0):
        raise RuntimeError("Passed in covariance matrix should be positive definite")
    if not scipy.linalg.issymmetric(cov):
        raise RuntimeError(f"Passed in covariance matrix {cov} should be symmetric")

    if len(mean) != d or len(cov) != d:
        raise RuntimeError(f"Dimensionality of mean and covariance matrix should match {d}")

    data = rng.multivariate_normal(mean=mean, cov=cov, size=(n_samples))

    return data, mean, cov



# cython: boundscheck=False
# cython: wraparound=False
# cython: initializedcheck=False

import numpy as np

cimport numpy as cnp

cnp.import_array()

from libc.math cimport isnan
from libc.stdlib cimport free, malloc
from libcpp.stack cimport stack

from ..._lib.sklearn.tree._tree cimport ParentInfo, _build_pruned_tree

TREE_LEAF = -1
cdef intp_t _TREE_LEAF = TREE_LEAF
cdef float64_t INFINITY = np.inf

cdef inline void _init_parent_record(ParentInfo* record) noexcept nogil:
    record.n_constant_features = 0
    record.lower_bound = -INFINITY
    record.upper_bound = INFINITY


def _build_pruned_tree_honesty(
    Tree tree,
    Tree orig_tree,
    HonestPruner pruner,
    object X,
    const float64_t[:, ::1] y,
    const float64_t[:] sample_weight,
    const uint8_t[::1] missing_values_in_feature_mask=None,
):
    """Prune an existing tree with honest splits.

    Parameters
    ----------
    tree : Tree
        The tree to be pruned.
    orig_tree : Tree
        The original tree to be pruned.
    pruner : HonestPruner
        The pruner to enforce honest splits.
    X : array-like of shape (n_samples, n_features)
        The input samples.
    y : array-like of shape (n_samples,)
        The target values.
    sample_weight : array-like of shape (n_samples,)
        The sample weights.
    missing_values_in_feature_mask : array-like of shape (n_features,)
        The mask of missing values in the features.
    """
    cdef:
        intp_t n_nodes = orig_tree.node_count
        uint8_t[:] leaves_in_subtree = np.zeros(
            shape=n_nodes, dtype=np.uint8)

    # initialize the pruner/splitter
    pruner.init(X, y, sample_weight, missing_values_in_feature_mask)

    # apply pruning to the tree
    _honest_prune(leaves_in_subtree, orig_tree, pruner)

    _build_pruned_tree(tree, orig_tree, leaves_in_subtree,
                       pruner.capacity)


cdef class HonestPruner(Splitter):
    """Pruning to enforce honest splits are non-degenerate."""

    def __cinit__(
        self,
        Criterion criterion,
        intp_t max_features,
        intp_t min_samples_leaf,
        float64_t min_weight_leaf,
        object random_state,
        const int8_t[:] monotonic_cst,
        Tree orig_tree,
        *argv
    ):
        """
        Parameters
        ----------
        criterion : Criterion
            The criterion to measure the quality of a split.

        max_features : intp_t
            The maximal number of randomly selected features which can be
            considered for a split.

        min_samples_leaf : intp_t
            The minimal number of samples each leaf can have, where splits
            which would result in having less samples in a leaf are not
            considered.

        min_weight_leaf : float64_t
            The minimal weight each leaf can have, where the weight is the sum
            of the weights of each sample in it.

        random_state : object
            The user inputted random state to be used for pseudo-randomness

        monotonic_cst : const int8_t[:]
            Monotonicity constraints

        orig_tree : Tree
            The original tree to be pruned.
        """
        self.tree = orig_tree
        self.capacity = 0

    cdef int init(
        self,
        object X,
        const float64_t[:, ::1] y,
        const float64_t[:] sample_weight,
        const uint8_t[::1] missing_values_in_feature_mask,
    ) except -1:
        Splitter.init(self, X, y, sample_weight, missing_values_in_feature_mask)
        self.X = X

    cdef int partition_samples(
        self,
        intp_t node_idx,
    ) noexcept nogil:
        """Partition samples for X at the threshold and feature index of `orig_tree`.

        If missing values are present, this method partitions `samples`
        so that the `best_n_missing` missing values' indices are in the
        right-most end of `samples`, that is `samples[end_non_missing:end]`.
        """
        cdef float64_t threshold = self.tree.nodes[node_idx].threshold
        cdef intp_t feature = self.tree.nodes[node_idx].feature
        cdef intp_t n_missing = 0
        cdef intp_t pos = self.start
        cdef intp_t p
        cdef intp_t sample_idx
        cdef intp_t current_end = self.end - n_missing
        cdef const float32_t[:, :] X_ndarray = self.X

        # partition the samples one by one by swapping them to the left or right
        # of pos depending on the feature value compared to the orig_tree threshold
        for p in range(self.start, self.end):
            sample_idx = self.samples[p]

            # missing-values are always placed at the right-most end
            if isnan(X_ndarray[sample_idx, feature]):
                self.samples[p], self.samples[current_end] = \
                    self.samples[current_end], self.samples[p]

                n_missing += 1
                current_end -= 1
            elif p > pos and (self.tree._compute_feature(X_ndarray, sample_idx, &self.tree.nodes[node_idx])<= threshold):
                self.samples[p], self.samples[pos] = \
                    self.samples[pos], self.samples[p]
                pos += 1

        # this is the split point for left/right children
        self.pos = pos
        self.n_missing = n_missing
        self.missing_go_to_left = self.tree.nodes[node_idx].missing_go_to_left

    cdef bint check_node_partition_conditions(
        self,
        SplitRecord* current_split,
        float64_t lower_bound,
        float64_t upper_bound
    ) noexcept nogil:
        """Check that the current node satisfies paritioning conditions.

        Parameters
        ----------
        current_split : SplitRecord pointer
            A pointer to a memory-allocated SplitRecord object which will be filled with the
            split chosen.
        """
        # update the criterion if we are checking split conditions
        self.criterion.init_missing(self.n_missing)
        self.criterion.reset()
        self.criterion.update(self.pos)

        current_split.pos = self.pos
        current_split.n_missing = self.n_missing
        current_split.missing_go_to_left = self.missing_go_to_left

        # first check the presplit conditions
        cdef bint invalid_split = self.check_presplit_conditions(
            current_split,
            self.n_missing,
            self.missing_go_to_left
        )

        if invalid_split:
            return 0

        # TODO: make work with lower/upper bound. This will require passing
        # lower/upper bound from the parent node into check_node_partition_conditions
        # Reject if monotonicity constraints are not satisfied
        if (
            self.with_monotonic_cst and
            self.monotonic_cst[current_split.feature] != 0 and
            not self.criterion.check_monotonicity(
                self.monotonic_cst[current_split.feature],
                lower_bound,
                upper_bound,
            )
        ):
            return 0

        # Note this is called after pre-split condition checks
        # shift missing values to left if required, so we can check
        # the split conditions
        shift_missing_values_to_left_if_required(
            current_split,
            self.samples,
            self.end
        )

        # next check the postsplit conditions that leverages the criterion
        invalid_split = self.check_postsplit_conditions()
        return invalid_split

    cdef inline intp_t n_left_samples(
        self
    ) noexcept nogil:
        """Number of samples to send to the left child."""
        cdef intp_t n_left

        if self.missing_go_to_left:
            n_left = self.pos - self.start + self.n_missing
        else:
            n_left = self.pos - self.start
        return n_left

    cdef inline intp_t n_right_samples(
        self
    ) noexcept nogil:
        """Number of samples to send to the right child."""
        cdef intp_t n_right
        cdef intp_t end_non_missing = self.end - self.n_missing
        if self.missing_go_to_left:
            n_right = end_non_missing - self.pos
        else:
            n_right = end_non_missing - self.pos + self.n_missing
        return n_right

    cdef int node_split(
        self,
        ParentInfo* parent_record,
        SplitRecord* split,
    ) except -1 nogil:
        """Split nodes using the already constructed tree.

        This is a simpler version of splitting nodes during the construction of a tree.
        Here, we only need to split the samples in the node based on the feature and
        threshold of the node in the original tree. In addition, we track the relevant
        information from the parent node, such as lower/upper bounds, and the parent's
        impurity, and n_constant_features.

        Returns 0 if a split cannot be done, 1 if a split can be done
        and -1 in case of failure to allocate memory (and raise MemoryError).
        """
        raise NotImplementedError("node_split is not used in honest pruning")


cdef _honest_prune(
    uint8_t[:] leaves_in_subtree,
    Tree orig_tree,
    HonestPruner pruner,
):
    """Perform honest pruning of the tree.

    Iterates through the original tree in a BFS fashion using the pruner
    and tracks at each node (orig_node):

    - the number of samples in the node
    - the number of samples that would be sent to the left and right children

    Until one of three stopping conditions are met:

    1. The orig_node is a leaf node in the original tree.
        Thus we keep the node as a leaf in the pruned tree.
    2. The orig_node is a non-leaf node and the split is degenerate.
        Thus we would prune the subtree, and assign orig_node as a leaf
        node in the pruned tree.
    3. Stopping criterion is met based on the samples that reach the node.
        These are the stopping conditions implemented in a Splitter/Pruner.
        Thus we would prune the subtree, and assign orig_node as a leaf
        node in the pruned tree.

    Parameters
    ----------
    leaves_in_subtree : array of shape (n_nodes,), dtype=np.uint8
        Array of booleans indicating whether the node is in the subtree.
    orig_tree : Tree
        The original tree.
    pruner : HonestPruner
        The input samples to be used for computing the split of samples
        in the nodes.
    """
    cdef:
        # get the left child, right child and parents of every node
        intp_t[:] child_l = orig_tree.children_left
        intp_t[:] child_r = orig_tree.children_right
        # intp_t[:] parents = np.zeros(shape=n_nodes, dtype=np.intp)

        # stack to keep track of the nodes to be pruned such that BFS is done
        stack[PruningRecord] pruning_stack
        PruningRecord stack_record

        intp_t node_idx
        SplitRecord* split_ptr = <SplitRecord *>malloc(pruner.pointer_size())

        bint is_leaf_in_origtree
        bint invalid_split
        bint split_is_degenerate

        intp_t start = 0
        intp_t end = 0
        float64_t weighted_n_node_samples

        float64_t lower_bound, upper_bound
        float64_t left_child_min, left_child_max, right_child_min, right_child_max, middle_value

    # find parent node ids and leaves
    with nogil:
        # Push the root node
        pruning_stack.push({
            "node_idx": 0,
            "start": 0,
            "end": pruner.n_samples,
            "lower_bound": -INFINITY,
            "upper_bound": INFINITY,
        })

        # Note: this DFS building strategy differs from scikit-learn in that
        # we check stopping conditions (and leaf candidacy) after a split occurs.
        # If we don't hit a leaf, then we will add the children to the stack, but otherwise
        # we will halt the split, and mark the node to be a new leaf node in the pruned tree.
        while not pruning_stack.empty():
            stack_record = pruning_stack.top()
            pruning_stack.pop()
            start = stack_record.start
            end = stack_record.end
            lower_bound = stack_record.lower_bound
            upper_bound = stack_record.upper_bound

            # node index of actual node within the orig_tree
            node_idx = stack_record.node_idx

            # reset which samples indices are considered at this split node
            pruner.node_reset(start, end, &weighted_n_node_samples)

            # partition samples into left/right child based on the
            # current node split in the orig_tree
            pruner.partition_samples(node_idx)

            # check end conditions
            split_ptr.feature = orig_tree.nodes[node_idx].feature
            invalid_split = pruner.check_node_partition_conditions(
                split_ptr,
                lower_bound,
                upper_bound
            )
            split_is_degenerate = (
                pruner.n_left_samples() == 0 or pruner.n_right_samples() == 0
            )
            is_leaf_in_origtree = child_l[node_idx] == _TREE_LEAF
            if invalid_split or split_is_degenerate or is_leaf_in_origtree:
                # ... and child_r[node_idx] == _TREE_LEAF:
                #
                # 1) if node is not degenerate, that means there are still honest-samples in
                # both left/right children of the proposed split, or the node itself is a leaf
                # or 2) there are still nodes to split on, but the honest-samples have been
                # used up so the "parent" should be the new leaf
                leaves_in_subtree[node_idx] = 1
            else:
                if (
                    not pruner.with_monotonic_cst or
                    pruner.monotonic_cst[split_ptr.feature] == 0
                ):
                    # Split on a feature with no monotonicity constraint

                    # Current bounds must always be propagated to both children.
                    # If a monotonic constraint is active, bounds are used in
                    # node value clipping.
                    left_child_min = right_child_min = lower_bound
                    left_child_max = right_child_max = upper_bound
                elif pruner.monotonic_cst[split_ptr.feature] == 1:
                    # Split on a feature with monotonic increase constraint
                    left_child_min = lower_bound
                    right_child_max = upper_bound

                    # Lower bound for right child and upper bound for left child
                    # are set to the same value.
                    middle_value = pruner.criterion.middle_value()
                    right_child_min = middle_value
                    left_child_max = middle_value
                else:  # i.e. pruner.monotonic_cst[split.feature] == -1
                    # Split on a feature with monotonic decrease constraint
                    right_child_min = lower_bound
                    left_child_max = upper_bound

                    # Lower bound for left child and upper bound for right child
                    # are set to the same value.
                    middle_value = pruner.criterion.middle_value()
                    left_child_min = middle_value
                    right_child_max = middle_value

                pruning_stack.push({
                    "node_idx": child_l[node_idx],
                    "start": pruner.start,
                    "end": pruner.pos,
                    "lower_bound": left_child_min,
                    "upper_bound": left_child_max,
                })
                pruning_stack.push({
                    "node_idx": child_r[node_idx],
                    "start": pruner.pos,
                    "end": pruner.end,
                    "lower_bound": right_child_min,
                    "upper_bound": right_child_max,
                })

    # free the memory created for the SplitRecord pointer
    free(split_ptr)



import numpy as np
import pytest
from sklearn.datasets import make_blobs, make_classification
from sklearn.ensemble import ExtraTreesClassifier, RandomForestClassifier
from sklearn.neighbors import NearestNeighbors
from sklearn.tree import (
    DecisionTreeClassifier,
    DecisionTreeRegressor,
    ExtraTreeClassifier,
    ExtraTreeRegressor,
)
from sklearn.utils.estimator_checks import parametrize_with_checks

from treeple.ensemble import (
    ObliqueRandomForestClassifier,
    PatchObliqueRandomForestClassifier,
    UnsupervisedObliqueRandomForest,
    UnsupervisedRandomForest,
)
from treeple.neighbors import NearestNeighborsMetaEstimator

FORESTS = [
    ObliqueRandomForestClassifier,
    PatchObliqueRandomForestClassifier,
    UnsupervisedRandomForest,
    UnsupervisedObliqueRandomForest,
]


@pytest.mark.parametrize("forest", FORESTS)
def test_similarity_matrix(forest):
    n_samples = 200
    n_classes = 2
    n_features = 5

    X, y = make_blobs(
        n_samples=n_samples, centers=n_classes, n_features=n_features, random_state=12345
    )

    clf = forest(random_state=12345)
    clf.fit(X, y)
    sim_mat = clf.compute_similarity_matrix(X)

    assert sim_mat.shape == (n_samples, n_samples)
    assert np.allclose(sim_mat, sim_mat.T)
    assert np.all((sim_mat.diagonal() == 1))


@pytest.fixture
def sample_data():
    # Generate sample data for testing
    X, y = make_classification(n_samples=100, n_features=10, random_state=42)
    return X, y


@pytest.mark.parametrize(
    "estimator",
    [
        DecisionTreeClassifier(random_state=0),
        DecisionTreeRegressor(random_state=0),
        ExtraTreeClassifier(random_state=0),
        ExtraTreeRegressor(random_state=0),
        RandomForestClassifier(random_state=0, n_estimators=10),
        ExtraTreesClassifier(random_state=0, n_estimators=10),
    ],
)
def test_nearest_neighbors_meta_estimator(sample_data, estimator):
    X, y = sample_data
    estimator.fit(X, y)

    meta_estimator = NearestNeighborsMetaEstimator(estimator)

    # Fit the meta-estimator
    meta_estimator.fit(X, y)

    # Test the fitted estimator attribute
    assert hasattr(meta_estimator, "estimator_")

    # Test the nearest neighbors estimator
    assert isinstance(meta_estimator.neigh_est_, NearestNeighbors)

    # Test the kneighbors method
    neigh_dist, neigh_ind = meta_estimator.kneighbors()
    assert neigh_dist.shape == (X.shape[0], meta_estimator.n_neighbors)
    assert neigh_ind.shape == (X.shape[0], meta_estimator.n_neighbors)

    # Test the radius_neighbors method
    neigh_dist, neigh_ind = meta_estimator.radius_neighbors(radius=0.5)
    assert neigh_dist.shape == (X.shape[0],)
    assert neigh_ind.shape == (X.shape[0],)


@parametrize_with_checks(
    [
        NearestNeighborsMetaEstimator(DecisionTreeClassifier(random_state=0)),
    ]
)
def test_sklearn_compatible_transformer(estimator, check):
    check(estimator)



"""Instructions
1. Build this PR and run:

```bash
python bench_morf.py bench ~/bench_results_forest pr
```

2. On main run:

```bash
python bench_morf.py bench ~/bench_results_forest main
```

3. Plotting

```bash
python bench_morf.py plot ~/bench_results_forest
    pr main results_image.png
```
"""

import argparse
import csv
from functools import partial
from itertools import product
from pathlib import Path
from statistics import mean, stdev
from time import perf_counter

import numpy as np
from sklearn.datasets import make_classification, make_low_rank_matrix, make_regression

from treeple import PatchObliqueRandomForestClassifier, PatchObliqueRandomForestRegressor


def make_poisson_data(n_samples, n_features=50, random_state=0, has_missing=False):
    rng = np.random.RandomState(random_state)
    X = make_low_rank_matrix(n_samples=n_samples, n_features=n_features, random_state=rng)
    coef = rng.uniform(low=-2, high=2, size=n_features) / np.max(X, axis=0)
    y = rng.poisson(lam=np.exp(X @ coef))
    # if has_missing:
    #     missing_mask = rng.choice([True, False], size=X.shape, p=[0.1, 0.9])
    #     X[missing_mask] = np.nan
    return X, y


def make_low_card_data(n_samples, n_features=50, random_state=0, has_missing=False):
    rng = np.random.RandomState(random_state)
    X = rng.choice([0.0, 1.0, 2.0], size=(n_samples, n_features))
    y = rng.choice([0, 1], size=n_samples)
    # if has_missing:
    #     missing_mask = rng.choice([True, False], size=X.shape, p=[0.1, 0.9])
    #     X[missing_mask] = np.nan
    return X, y


def make_regression_custom(*args, has_missing=False, random_state=None, **kwargs):
    X, y = make_regression(*args, random_state=random_state, **kwargs)
    rng = np.random.RandomState()
    # if has_missing:
    #     rng = np.random.RandomState(random_state)
    #     missing_mask = rng.choice([True, False], size=X.shape, p=[0.1, 0.9])
    #     X[missing_mask] = np.nan
    return X, y


def make_classification_custom(*args, has_missing=False, random_state=None, **kwargs):
    X, y = make_classification(*args, random_state=random_state, **kwargs)
    rng = np.random.RandomState()
    # if has_missing:
    #     rng = np.random.RandomState(random_state)
    #     missing_mask = rng.choice([True, False], size=X.shape, p=[0.1, 0.9])
    #     X[missing_mask] = np.nan
    return X, y


N_REPEATS = 10

benchmark_config = [
    (
        PatchObliqueRandomForestRegressor,
        list(
            product(
                ["squared_error"],
                [
                    partial(make_regression_custom, n_targets=2),
                    make_low_card_data,
                ],
                [10_000],
                ["dense"],
                ["best"],
                [False],
            )
        ),
    ),
    (
        PatchObliqueRandomForestClassifier,
        list(
            product(
                ["gini", "entropy"],
                [
                    partial(make_classification_custom, n_informative=100, n_classes=5),
                    make_low_card_data,
                ],
                [10_000],
                ["dense"],
                ["best"],
                [False],
            )
        ),
    ),
]


def bench(args):
    bench_results, branch = args.bench_results, args.branch
    results_dir = Path(bench_results)
    results_dir.mkdir(exist_ok=True)

    results_path = results_dir / f"{branch}.csv"

    with results_path.open("w") as f:
        writer = csv.DictWriter(
            f,
            fieldnames=[
                "criterion",
                "n_samples",
                "make_data",
                "container",
                "splitter",
                "n_repeat",
                "duration",
            ],
        )
        writer.writeheader()

        for Klass, items in benchmark_config:
            for config in items:
                (
                    criterion,
                    make_data,
                    n_samples,
                    container,
                    splitter,
                ) = config
                if isinstance(make_data, partial):
                    make_data_str = make_data.func.__name__
                else:
                    make_data_str = make_data.__name__

                default_config = {
                    "criterion": criterion,
                    "n_samples": n_samples,
                    "make_data": make_data_str,
                    "container": container,
                    "splitter": splitter,
                }
                combine_config = " ".join(f"{k}={v}" for k, v in default_config.items())

                klass_results = []
                for n_repeat in range(N_REPEATS):
                    X, y = make_data(
                        n_samples=n_samples,
                        n_features=20,
                        random_state=n_repeat,
                        has_missing=False,
                    )
                    tree = Klass(random_state=n_repeat, criterion=criterion)

                    start = perf_counter()
                    tree.fit(X, y)
                    duration = perf_counter() - start
                    klass_results.append(duration)
                    writer.writerow(
                        {
                            **default_config,
                            **{
                                "n_repeat": n_repeat,
                                "duration": duration,
                            },
                        }
                    )
                results_mean, results_stdev = mean(klass_results), stdev(klass_results)
                print(f"{combine_config} with {results_mean:.3f} +/- {results_stdev:.3f}")


def plot(args):
    import matplotlib.pyplot as plt
    import pandas as pd
    import seaborn as sns

    results_path = Path(args.bench_results)
    pr_path = results_path / f"{args.pr_name}.csv"
    main_path = results_path / f"{args.main_name}.csv"
    image_path = results_path / args.image_path

    df_pr = pd.read_csv(pr_path).assign(branch=args.pr_name)
    df_main = pd.read_csv(main_path).assign(branch=args.main_name)
    df_all = pd.concat((df_pr, df_main), ignore_index=True)

    df_all = df_all.assign(
        make_data=df_all["make_data"]
        .str.replace("_custom", "")
        .str.replace("make_", "")
        .str.replace("_data", "")
    )

    gb = df_all.groupby(["criterion", "make_data"])
    groups = gb.groups

    n_rows, n_cols = 2, 4
    fig, axes = plt.subplots(n_rows, n_cols, figsize=(12, 8), constrained_layout=True)
    axes_flat = axes.ravel()
    for i, (keys, idx) in enumerate(groups.items()):
        ax = axes_flat[i]
        ax.set_title(" | ".join(keys))
        sns.boxplot(data=df_all.loc[idx], y="duration", x="branch", ax=ax)
        if i % n_cols != 0:
            ax.set_ylabel("")

    axes_flat[-1].set_visible(False)

    fig.savefig(image_path)
    print(f"Saved image to {image_path}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    subparsers = parser.add_subparsers()
    bench_parser = subparsers.add_parser("bench")
    bench_parser.add_argument("bench_results")
    bench_parser.add_argument("branch")
    bench_parser.set_defaults(func=bench)

    plot_parser = subparsers.add_parser("plot")
    plot_parser.add_argument("bench_results")
    plot_parser.add_argument("pr_name")
    plot_parser.add_argument("main_name")
    plot_parser.add_argument("image_path")
    plot_parser.set_defaults(func=plot)

    args = parser.parse_args()
    args.func(args)



import numpy as np
import pytest
from numpy.testing import assert_array_equal
from sklearn.datasets import make_blobs
from sklearn.metrics import accuracy_score
from sklearn.model_selection import cross_val_score, train_test_split
from sklearn.utils.estimator_checks import parametrize_with_checks

from treeple import MultiViewRandomForestClassifier, RandomForestClassifier
from treeple.datasets.multiview import make_joint_factor_model

seed = 12345


@parametrize_with_checks(
    [
        MultiViewRandomForestClassifier(random_state=12345, n_estimators=10),
    ]
)
def test_sklearn_compatible_estimator(estimator, check):
    if check.func.__name__ in [
        # sample weights do not necessarily imply a sample is not used in clustering
        "check_sample_weight_equivalence",
        "check_sample_weight_equivalence_on_dense_data",
        "check_sample_weight_equivalence_on_sparse_data",
    ]:
        pytest.skip()
    check(estimator)


@pytest.mark.parametrize("baseline_est", [RandomForestClassifier])
def test_multiview_classification(baseline_est):
    """Test that explicit knowledge of multi-view structure improves classification accuracy.

    In very high-dimensional noise setting across two views, when the max_depth and max_features
    are constrained, the multi-view random forests will still obtain perfect accuracy, while
    the single-view random forest will not.
    """
    rng = np.random.default_rng(seed=seed)

    n_samples = 100
    n_estimators = 10
    n_features_1 = 5
    n_features_2 = 1000
    cluster_std = 2.0

    # Create a high-dimensional multiview dataset with a low-dimensional informative
    # subspace in one view of the dataset.
    X0_first, y0 = make_blobs(
        n_samples=n_samples,
        cluster_std=cluster_std,
        n_features=n_features_1,
        random_state=rng.integers(1, 10000),
        centers=1,
    )

    X1_first, y1 = make_blobs(
        n_samples=n_samples,
        cluster_std=cluster_std,
        n_features=n_features_1,
        random_state=rng.integers(1, 10000),
        centers=1,
    )
    y1[:] = 1
    X0 = np.concatenate([X0_first, rng.standard_normal(size=(n_samples, n_features_2))], axis=1)
    X1 = np.concatenate([X1_first, rng.standard_normal(size=(n_samples, n_features_2))], axis=1)
    X = np.vstack((X0, X1))
    y = np.hstack((y0, y1)).T

    # Compare multiview decision tree vs single-view decision tree
    clf = MultiViewRandomForestClassifier(
        random_state=seed,
        feature_set_ends=[n_features_1, X.shape[1]],
        max_features="sqrt",
        max_depth=4,
        n_estimators=n_estimators,
    )
    clf.fit(X, y)
    assert (
        accuracy_score(y, clf.predict(X)) == 1.0
    ), f"Accuracy score: {accuracy_score(y, clf.predict(X))}"
    assert (
        cross_val_score(clf, X, y, cv=5).mean() == 1.0
    ), f"CV score: {cross_val_score(clf, X, y, cv=5).mean()}"

    clf = baseline_est(
        random_state=seed,
        max_depth=4,
        max_features="sqrt",
        n_estimators=n_estimators,
    )
    clf.fit(X, y)
    assert (
        accuracy_score(y, clf.predict(X)) > 0.99
    ), f"Accuracy score: {accuracy_score(y, clf.predict(X))}"
    assert (
        cross_val_score(clf, X, y, cv=3).mean() <= 0.95
    ), f"CV score for {clf}: {cross_val_score(clf, X, y, cv=3).mean()}"


@pytest.mark.parametrize(
    "n_views, max_features",
    [
        (2, 0.3),
        (3, 0.3),
        (4, 0.3),
        (5, 0.3),
    ],
)
def test_three_view_dataset(n_views, max_features):
    rng = np.random.default_rng(seed=seed)

    n_estimators = 50

    n_samples = 100
    n_views = 5
    n_features = 200

    # number of dimensions that are informative
    n_signals = 5

    # Create a high-dimensional multiview dataset
    Xs = make_joint_factor_model(
        n_features=n_signals,
        m=2.0,
        n_views=n_views,
        n_samples=n_samples,
        joint_rank=5,
        random_state=seed + 1,
    )
    X = np.empty((n_samples, 0))
    feature_set_ends = []
    for _X in Xs:
        X = np.hstack((X, _X, rng.standard_normal(size=(n_samples, n_features - n_signals))))
        feature_set_ends.append(X.shape[1])

    Xs = make_joint_factor_model(
        n_features=n_signals,
        m=2.0,
        n_views=n_views,
        n_samples=n_samples,
        joint_rank=5,
        random_state=seed + 2,
    )
    second_X = np.empty((n_samples, 0))
    for _X in Xs:
        second_X = np.hstack(
            (second_X, _X, rng.standard_normal(size=(n_samples, n_features - n_signals)))
        )
    X = np.vstack((X, second_X))
    y = np.concatenate([np.zeros(n_samples), np.ones(n_samples)]).T

    # perform train/test split
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=seed)

    # Compare multiview decision tree vs single-view decision tree
    clf = MultiViewRandomForestClassifier(
        random_state=seed,
        feature_set_ends=feature_set_ends,
        apply_max_features_per_feature_set=True,
        max_features=max_features,
        n_estimators=n_estimators,
    )
    clf.fit(X_train, y_train)

    rf_clf = RandomForestClassifier(
        random_state=seed,
        max_features=max_features,
        n_estimators=n_estimators,
    )
    rf_clf.fit(X_train, y_train)
    print(n_features, max_features, n_features * max_features)
    assert_array_equal(
        clf.estimators_[0].max_features_per_set_, [n_features * max_features] * n_views
    )

    rf_acc = accuracy_score(y_test, rf_clf.predict(X_test))
    acc = accuracy_score(y_test, clf.predict(X_test))
    assert acc > rf_acc + 0.01, f"Accuracy score: {acc} < {rf_acc}"



import numpy as np


def compute_forest_similarity_matrix(forest, X):
    """Compute the similarity matrix of samples in X using a trained forest.

    As an intermediate calculation, the forest applies the dataset and gets
    the leaves for each sample. Then, the similarity matrix is computed by
    counting the number of times each pair of samples ends up in the same leaf.

    Parameters
    ----------
    forest : BaseForest or BaseDecisionTree
        The fitted forest.
    X : array-like of shape (n_samples, n_features)
        The input data.

    Returns
    -------
    aff_matrix : array-like of shape (n_samples, n_samples)
        The estimated distance matrix.
    """
    if hasattr(forest, "estimator_"):
        # apply to the leaves
        X_leaves = forest.apply(X)

        n_est = forest.n_estimators
    else:
        # apply to the leaves for a single tree
        X_leaves = forest.apply(X)[:, np.newaxis]
        n_est = 1

    aff_matrix = sum(np.equal.outer(X_leaves[:, i], X_leaves[:, i]) for i in range(n_est))
    # normalize by the number of trees
    aff_matrix = np.divide(aff_matrix, n_est)
    return aff_matrix


def _compute_distance_matrix(aff_matrix):
    """Private function to compute distance matrix after `compute_similarity_matrix`."""
    dists = 1.0 - aff_matrix
    return dists


# ported from https://github.com/neurodata/hyppo/blob/main/hyppo/independence/_utils.py
class SimMatrixMixin:
    """Mixin class to calculate similarity and dissimilarity matrices.

    This augments tree/forest models with the sklearn's nearest-neighbors API.
    """

    def compute_similarity_matrix(self, X):
        """
        Compute the similarity matrix of samples in X.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input data.

        Returns
        -------
        sim_matrix : array-like of shape (n_samples, n_samples)
            The similarity matrix among the samples.
        """
        return compute_forest_similarity_matrix(self, X)



from sklearn.utils._param_validation import StrOptions

from .._lib.sklearn.ensemble._forest import ForestClassifier
from ..tree import MultiViewDecisionTreeClassifier
from ..tree._neighbors import SimMatrixMixin
from ._extensions import ForestClassifierMixin, ForestMixin


class MultiViewRandomForestClassifier(
    SimMatrixMixin, ForestClassifierMixin, ForestMixin, ForestClassifier
):
    """
    A multi-view axis-aligned random forest classifier.

    A multi-view random forest is a meta estimator similar to a random
    forest that fits a number of multi-view decision tree classifiers
    on various sub-samples of the dataset and uses averaging to
    improve the predictive accuracy and control over-fitting.

    Parameters
    ----------
    n_estimators : int, default=100
        The number of trees in the forest.

    criterion : {"gini", "entropy"}, default="gini"
        The function to measure the quality of a split. Supported criteria are
        "gini" for the Gini impurity and "entropy" for the information gain.
        Note: this parameter is tree-specific.

    max_depth : int, default=None
        The maximum depth of the tree. If None, then nodes are expanded until
        all leaves are pure or until all leaves contain less than
        min_samples_split samples.

    min_samples_split : int or float, default=2
        The minimum number of samples required to split an internal node:

        - If int, then consider `min_samples_split` as the minimum number.
        - If float, then `min_samples_split` is a fraction and
          `ceil(min_samples_split * n_samples)` are the minimum
          number of samples for each split.

    min_samples_leaf : int or float, default=1
        The minimum number of samples required to be at a leaf node.
        A split point at any depth will only be considered if it leaves at
        least ``min_samples_leaf`` training samples in each of the left and
        right branches.  This may have the effect of smoothing the model,
        especially in regression.

        - If int, then consider `min_samples_leaf` as the minimum number.
        - If float, then `min_samples_leaf` is a fraction and
          `ceil(min_samples_leaf * n_samples)` are the minimum
          number of samples for each node.

    min_weight_fraction_leaf : float, default=0.0
        The minimum weighted fraction of the sum total of weights (of all
        the input samples) required to be at a leaf node. Samples have
        equal weight when sample_weight is not provided.

    max_features : {"sqrt", "log2", None}, int or float, default="sqrt"
        The number of features to consider when looking for the best split:

        - If int, then consider `max_features` features at each split.
        - If float, then `max_features` is a fraction and
          `round(max_features * n_features)` features are considered at each
          split.
        - If "auto", then `max_features=sqrt(n_features)`.
        - If "sqrt", then `max_features=sqrt(n_features)`.
        - If "log2", then `max_features=log2(n_features)`.
        - If None, then `max_features=n_features`.

        Note: the search for a split does not stop until at least one
        valid partition of the node samples is found, even if it requires to
        effectively inspect more than ``max_features`` features.

    max_leaf_nodes : int, default=None
        Grow trees with ``max_leaf_nodes`` in best-first fashion.
        Best nodes are defined as relative reduction in impurity.
        If None then unlimited number of leaf nodes.

    min_impurity_decrease : float, default=0.0
        A node will be split if this split induces a decrease of the impurity
        greater than or equal to this value.

        The weighted impurity decrease equation is the following::

            N_t / N * (impurity - N_t_R / N_t * right_impurity
                                - N_t_L / N_t * left_impurity)

        where ``N`` is the total number of samples, ``N_t`` is the number of
        samples at the current node, ``N_t_L`` is the number of samples in the
        left child, and ``N_t_R`` is the number of samples in the right child.

        ``N``, ``N_t``, ``N_t_R`` and ``N_t_L`` all refer to the weighted sum,
        if ``sample_weight`` is passed.

    bootstrap : bool, default=True
        Whether bootstrap samples are used when building trees. If False, the
        whole dataset is used to build each tree.

    oob_score : bool, default=False
        Whether to use out-of-bag samples to estimate the generalization score.
        Only available if bootstrap=True.

    n_jobs : int, default=None
        The number of jobs to run in parallel. :meth:`fit`, :meth:`predict`,
        :meth:`decision_path` and :meth:`apply` are all parallelized over the
        trees. ``None`` means 1 unless in a `joblib.parallel_backend`
        context. ``-1`` means using all processors. See :term:`Glossary
        <n_jobs>` for more details.

    random_state : int, RandomState instance or None, default=None
        Controls both the randomness of the bootstrapping of the samples used
        when building trees (if ``bootstrap=True``) and the sampling of the
        features to consider when looking for the best split at each node
        (if ``max_features < n_features``).
        See :term:`Glossary <random_state>` for details.

    verbose : int, default=0
        Controls the verbosity when fitting and predicting.

    warm_start : bool, default=False
        When set to ``True``, reuse the solution of the previous call to fit
        and add more estimators to the ensemble, otherwise, just fit a whole
        new forest. See :term:`the Glossary <warm_start>`.

    class_weight : {"balanced", "balanced_subsample"}, dict or list of dicts, \
            default=None
        Weights associated with classes in the form ``{class_label: weight}``.
        If not given, all classes are supposed to have weight one. For
        multi-output problems, a list of dicts can be provided in the same
        order as the columns of y.

        Note that for multioutput (including multilabel) weights should be
        defined for each class of every column in its own dict. For example,
        for four-class multilabel classification weights should be
        [{0: 1, 1: 1}, {0: 1, 1: 5}, {0: 1, 1: 1}, {0: 1, 1: 1}] instead of
        [{1:1}, {2:5}, {3:1}, {4:1}].

        The "balanced" mode uses the values of y to automatically adjust
        weights inversely proportional to class frequencies in the input data
        as ``n_samples / (n_classes * np.bincount(y))``

        The "balanced_subsample" mode is the same as "balanced" except that
        weights are computed based on the bootstrap sample for every tree
        grown.

        For multi-output, the weights of each column of y will be multiplied.

        Note that these weights will be multiplied with sample_weight (passed
        through the fit method) if sample_weight is specified.

    max_samples : int or float, default=None
        If bootstrap is True, the number of samples to draw from X
        to train each base estimator.

        - If None (default), then draw `X.shape[0]` samples.
        - If int, then draw `max_samples` samples.
        - If float, then draw `max_samples * X.shape[0]` samples. Thus,
          `max_samples` should be in the interval `(0.0, 1.0]`.

    feature_combinations : float, default=None
        The number of features to combine on average at each split
        of the decision trees. If ``None``, then will default to the minimum of
        ``(1.5, n_features)``. This controls the number of non-zeros is the
        projection matrix. Setting the value to 1.0 is equivalent to a
        traditional decision-tree. ``feature_combinations * max_features``
        gives the number of expected non-zeros in the projection matrix of shape
        ``(max_features, n_features)``. Thus this value must always be less than
        ``n_features`` in order to be valid.

    feature_set_ends : array-like of int of shape (n_feature_sets,), default=None
        The indices of the end of each feature set. For example, if the first
        feature set is the first 10 features, and the second feature set is the
        next 20 features, then ``feature_set_ends = [10, 30]``. If ``None``,
        then this will assume that there is only one feature set.

    apply_max_features_per_feature_set : bool, default=False
        Whether to apply sampling per feature set, where ``max_features`` is applied
        to each feature-set. If ``False``, then sampling
        is applied over the entire feature space.

    Attributes
    ----------
    estimators_ : list of treeple.tree.ObliqueDecisionTreeClassifier
        The collection of fitted sub-estimators.

    classes_ : ndarray of shape (n_classes,) or a list of such arrays
        The classes labels (single output problem), or a list of arrays of
        class labels (multi-output problem).

    n_classes_ : int or list
        The number of classes (single output problem), or a list containing the
        number of classes for each output (multi-output problem).

    n_features_ : int
        The number of features when ``fit`` is performed.

    n_features_in_ : int
        Number of features seen during :term:`fit`.

    feature_names_in_ : ndarray of shape (`n_features_in_`,)
        Names of features seen during :term:`fit`. Defined only when `X`
        has feature names that are all strings.

    n_outputs_ : int
        The number of outputs when ``fit`` is performed.

    feature_importances_ : ndarray of shape (n_features,)
        The impurity-based feature importances.
        The higher, the more important the feature.
        The importance of a feature is computed as the (normalized)
        total reduction of the criterion brought by that feature.  It is also
        known as the Gini importance.

        Warning: impurity-based feature importances can be misleading for
        high cardinality features (many unique values). See
        :func:`sklearn.inspection.permutation_importance` as an alternative.

    oob_score_ : float
        Score of the training dataset obtained using an out-of-bag estimate.
        This attribute exists only when ``oob_score`` is True.

    oob_decision_function_ : ndarray of shape (n_samples, n_classes) or \
            (n_samples, n_classes, n_outputs)
        Decision function computed with out-of-bag estimate on the training
        set. If n_estimators is small it might be possible that a data point
        was never left out during the bootstrap. In this case,
        `oob_decision_function_` might contain NaN. This attribute exists
        only when ``oob_score`` is True.

    See Also
    --------
    treeple.tree.ObliqueDecisionTreeClassifier : An oblique decision
        tree classifier.
    sklearn.ensemble.RandomForestClassifier : An axis-aligned decision
        forest classifier.
    """

    tree_type = "oblique"
    _parameter_constraints: dict = {
        **ForestClassifier._parameter_constraints,
        **MultiViewDecisionTreeClassifier._parameter_constraints,
        "class_weight": [
            StrOptions({"balanced_subsample", "balanced"}),
            dict,
            list,
            None,
        ],
    }
    _parameter_constraints.pop("splitter")

    def __init__(
        self,
        n_estimators=100,
        *,
        criterion="gini",
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_weight_fraction_leaf=0.0,
        max_features="sqrt",
        max_leaf_nodes=None,
        min_impurity_decrease=0.0,
        bootstrap=True,
        oob_score=False,
        n_jobs=None,
        random_state=None,
        verbose=0,
        warm_start=False,
        class_weight=None,
        max_samples=None,
        feature_combinations=None,
        feature_set_ends=None,
        apply_max_features_per_feature_set=False,
    ):
        super().__init__(
            estimator=MultiViewDecisionTreeClassifier(),
            n_estimators=n_estimators,
            estimator_params=(
                "criterion",
                "max_depth",
                "min_samples_split",
                "min_samples_leaf",
                "min_weight_fraction_leaf",
                "max_features",
                "max_leaf_nodes",
                "min_impurity_decrease",
                "random_state",
                "feature_combinations",
                "feature_set_ends",
                "apply_max_features_per_feature_set",
            ),
            bootstrap=bootstrap,
            oob_score=oob_score,
            n_jobs=n_jobs,
            random_state=random_state,
            verbose=verbose,
            warm_start=warm_start,
            class_weight=class_weight,
            max_samples=max_samples,
        )
        self.criterion = criterion
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.max_features = max_features
        self.feature_combinations = feature_combinations
        self.feature_set_ends = feature_set_ends
        self.apply_max_features_per_feature_set = apply_max_features_per_feature_set

        # unused by oblique forests
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease



from typing import Optional

import numpy as np
import scipy.linalg
import scipy.special
import scipy.stats
from numpy.typing import ArrayLike
from sklearn.neighbors import NearestNeighbors
from sklearn.preprocessing import StandardScaler

from treeple.ensemble import UnsupervisedObliqueRandomForest
from treeple.tree import compute_forest_similarity_matrix


def entropy_gaussian(cov):
    """Compute entropy of a multivariate Gaussian.

    Computes the analytical solution due to :footcite:`Darbellay1999Entropy`.

    Parameters
    ----------
    cov : array-like of shape (d,d)
        The covariance matrix of the distribution.

    Returns
    -------
    true_mi : float
        The true analytical mutual information of the generated multivariate Gaussian distribution.

    Notes
    -----
    The analytical solution for the true mutual information, ``I(X; Y)`` is given by::

        I(X; Y) = H(X) + H(Y) - H(X, Y) = -\\frac{1}{2} log(det(C))

    References
    ----------
    .. footbibliography::
    """
    d = cov.shape[0]

    true_entropy = 0.5 * d * (1 + np.log(2 * np.pi)) + 0.5 * np.log(np.linalg.det(cov))
    return true_entropy


def mi_gaussian(cov):
    """Compute mutual information of a multivariate Gaussian.

    Parameters
    ----------
    cov : array-like of shape (d,d)
        The covariance matrix of the distribution.

    Returns
    -------
    true_mi : float
        The true analytical entropy of the generated multivariate Gaussian distribution.

    Notes
    -----
    Analytical solution for entropy, :math:`H(X)` of a multivariate Gaussian is given by::

        H(X) = \\frac{d}{2} (1 + log(2\\pi)) + \\frac{1}{2} log(det(C))
    """
    # computes the true MI
    true_mi = -0.5 * np.log(np.linalg.det(cov))
    return true_mi


def cmi_gaussian(cov, x_index, y_index, z_index):
    """Compute the analytical CMI for a multivariate Gaussian distribution.

    Parameters
    ----------
    cov : array-like of shape (d,d)
        The covariance matrix of the distribution.
    x_index : list or int
        List of indices in ``cov`` that are for the X variable.
    y_index : list or int
        List of indices in ``cov`` that are for the Y variable.
    z_index : list or int
        List of indices in ``cov`` that are for the Z variable.

    Returns
    -------
    true_mi : float
        The true analytical mutual information of the generated multivariate Gaussian distribution.

    Notes
    -----
    Analytical solution for conditional mutual information, :math:`I(X;Y|Z)` of a
    multivariate Gaussian is given by::

        I(X;Y | Z) = H(X, Z) + H(Y, Z) - H(Z) - H(X, Y, Z)

    where we plug in the analytical solutions for entropy as shown in :func:`entropy_gaussian`.
    """
    x_index = np.atleast_1d(x_index)
    y_index = np.atleast_1d(y_index)
    z_index = np.atleast_1d(z_index)

    xz_index = np.concatenate((x_index, z_index)).squeeze()
    yz_index = np.concatenate((y_index, z_index)).squeeze()

    cov_xz = cov[xz_index, xz_index]
    cov_yz = cov[yz_index, yz_index]
    cov_z = cov[z_index, z_index]

    cmi = (
        entropy_gaussian(cov_xz)
        + entropy_gaussian(cov_yz)
        - entropy_gaussian(cov_z)
        - entropy_gaussian(cov)
    )
    return cmi


def entropy_weibull(alpha, k):
    """Analytical solution for entropy of Weibull distribution.

    https://en.wikipedia.org/wiki/Weibull_distribution
    """
    return np.euler_gamma * (1.0 - 1.0 / alpha) - np.log(alpha * np.power(k, 1.0 / alpha)) + 1


def mi_gamma(theta):
    """Analytical solution for"""
    return scipy.special.digamma(theta + 1) - np.log(theta)


def mi_from_entropy(hx, hy, hxy):
    """Analytic formula for MI given plug-in estimates of entropies."""
    return hx + hy - hxy


def cmi_from_entropy(hxz, hyz, hz, hxyz):
    """Analytic formula for CMI given plug-in estimates of entropies."""
    return hxz + hyz - hz - hxyz


def mutual_info_ksg(
    X,
    Y,
    Z=None,
    k: float = 0.2,
    metric="forest",
    algorithm="kd_tree",
    n_jobs: int = -1,
    transform: str = "rank",
    random_seed: Optional[int] = None,
):
    """Compute the generalized (conditional) mutual information KSG estimate.

    Parameters
    ----------
    X : ArrayLike of shape (n_samples, n_features_x)
        The X covariate space.
    Y : ArrayLike of shape (n_samples, n_features_y)
        The Y covariate space.
    Z : ArrayLike of shape (n_samples, n_features_z), optional
        The Z covariate space, by default None. If None, then the MI is computed.
        If Z is defined, then the CMI is computed.
    k : float, optional
        The number of neighbors to use in defining the radius, by default 0.2.
    metric : str
        Any distance metric accepted by :class:`sklearn.neighbors.NearestNeighbors`.
        If 'forest' (default), then uses an
        :class:`treeple.UnsupervisedObliqueRandomForest` to compute geodesic distances.
    algorithm : str, optional
        Method to use, by default 'knn'. Can be ('ball_tree', 'kd_tree', 'brute').
    n_jobs : int, optional
        Number of parallel jobs, by default -1.
    transform : one of {'rank', 'standardize', 'uniform'}
        Preprocessing, by default "rank".
    random_seed : int, optional
        Random seed, by default None.

    Returns
    -------
    val : float
        The estimated MI, or CMI value.

    Notes
    -----
    Given a dataset with ``n`` samples, the KSG estimator proceeds by:

    1. For fixed k, get the distance to the kth nearest-nbr in XYZ subspace, call it 'r'
    2. Get the number of NN in XZ subspace within radius 'r'
    3. Get the number of NN in YZ subspace within radius 'r'
    4. Get the number of NN in Z subspace within radius 'r'
    5. Apply analytic solution for KSG estimate

    For MI, the analytical solution is:

    .. math::

        \\psi(k) - E[(\\psi(n_x) + \\psi(n_y))] + \\psi(n)

    For CMI, the analytical solution is:

    .. math::

        \\psi(k) - E[(\\psi(n_{xz}) + \\psi(n_{yz}) - \\psi(n_{z}))]

    where :math:`\\psi` is the DiGamma function, and each expectation term
    is estimated by taking the sample average.

    Note that the :math:`n_i` terms denote the number of neighbors within
    radius 'r' in the subspace of 'i', where 'i' could be for example the
    X, Y, XZ, etc. subspaces. This term does not include the sample itself.
    """
    rng = np.random.default_rng(random_seed)

    data = np.hstack((X, Y))
    if Z is not None:
        data = np.hstack((data, Z))

    data = _preprocess_data(data, transform, rng)
    n_samples = data.shape[0]

    if k < 1:
        knn_here = max(1, int(k * n_samples))
    else:
        knn_here = max(1, int(k))

    if Z is not None:
        val = _cmi_ksg(data, X, Y, Z, metric, algorithm, knn_here, n_jobs)
    else:
        val = _mi_ksg(data, X, Y, metric, algorithm, knn_here, n_jobs)
    return val


def _preprocess_data(data, transform, rng):
    n_samples, n_features = data.shape

    # add minor noise to make sure there are no ties
    random_noise = rng.random((n_samples, n_features))
    data += 1e-5 * random_noise @ np.std(data, axis=0).reshape(n_features, 1)

    if transform == "standardize":
        # standardize with standard scaling
        data = data.astype(np.float64)
        scaler = StandardScaler()
        data = scaler.fit_transform(data)
    elif transform == "uniform":
        data = _trafo2uniform(data)
    elif transform == "rank":
        # rank transform each column
        data = scipy.stats.rankdata(data, axis=0)
    return data


def _mi_ksg(data, X, Y, metric, algorithm, knn_here, n_jobs):
    """Compute KSG estimate of MI."""
    n_samples = X.shape[0]

    # estimate distance to the kth NN in XYZ subspace for each sample
    # - get the radius we want to use per sample as the distance to the kth neighbor
    #   in the joint distribution space
    neigh = _compute_nn(data, algorithm=algorithm, metric=metric, k=knn_here, n_jobs=n_jobs)
    dists, _ = neigh.kneighbors()
    radius_per_sample = dists[:, -1]

    # compute on the subspace of X
    num_nn_x = _compute_radius_nbrs(
        X,
        radius_per_sample,
        knn_here,
        algorithm=algorithm,
        metric=metric,
        n_jobs=n_jobs,
    )

    # compute on the subspace of Y
    num_nn_y = _compute_radius_nbrs(
        Y,
        radius_per_sample,
        knn_here,
        algorithm=algorithm,
        metric=metric,
        n_jobs=n_jobs,
    )

    # compute the final MI value
    # \\psi(k) - E[(\\psi(n_x) + \\psi(n_y))] + \\psi(n)
    hxy = scipy.special.digamma(knn_here)
    hx = scipy.special.digamma(num_nn_x)
    hy = scipy.special.digamma(num_nn_y)
    hn = scipy.special.digamma(n_samples)
    val = hxy - (hx + hy).mean() + hn
    return val


def _cmi_ksg(data, X, Y, Z, metric, algorithm, knn_here, n_jobs):
    """Compute KSG estimate of CMI."""
    # estimate distance to the kth NN in XYZ subspace for each sample
    neigh = _compute_nn(data, algorithm=algorithm, metric=metric, k=knn_here, n_jobs=n_jobs)

    # get the radius we want to use per sample as the distance to the kth neighbor
    # in the joint distribution space
    dists, _ = neigh.kneighbors()
    radius_per_sample = dists[:, -1]

    # compute on the subspace of XZ
    xz_data = np.hstack((X, Z))
    num_nn_xz = _compute_radius_nbrs(
        xz_data,
        radius_per_sample,
        knn_here,
        algorithm=algorithm,
        metric=metric,
        n_jobs=n_jobs,
    )

    # compute on the subspace of YZ
    yz_data = np.hstack((Y, Z))
    num_nn_yz = _compute_radius_nbrs(
        yz_data,
        radius_per_sample,
        knn_here,
        algorithm=algorithm,
        metric=metric,
        n_jobs=n_jobs,
    )

    # compute on the subspace of XZ
    num_nn_z = _compute_radius_nbrs(
        Z,
        radius_per_sample,
        knn_here,
        algorithm=algorithm,
        metric=metric,
        n_jobs=n_jobs,
    )

    # compute the final CMI value
    hxyz = scipy.special.digamma(knn_here)
    hxz = scipy.special.digamma(num_nn_xz)
    hyz = scipy.special.digamma(num_nn_yz)
    hz = scipy.special.digamma(num_nn_z)
    val = hxyz - (hxz + hyz - hz).mean()
    return val


def _compute_radius_nbrs(
    data,
    radius_per_sample,
    k,
    algorithm: str = "kd_tree",
    metric="l2",
    n_jobs: Optional[int] = None,
):
    neigh = _compute_nn(data, algorithm=algorithm, metric=metric, k=k, n_jobs=n_jobs)

    n_samples = radius_per_sample.shape[0]

    num_nn_data = np.zeros((n_samples,))
    for idx in range(n_samples):
        nn = neigh.radius_neighbors(radius=radius_per_sample[idx], return_distance=False)
        num_nn_data[idx] = len(nn)
    return num_nn_data


def _compute_nn(
    X: ArrayLike, algorithm: str = "kd_tree", metric="l2", k: int = 1, n_jobs: Optional[int] = None
) -> NearestNeighbors:
    """Compute kNN in subspace.

    Parameters
    ----------
    X : ArrayLike of shape (n_samples, n_features)
        The covariate space.
    algorithm : str, optional
        Method to use, by default 'knn'. Can be ('ball_tree', 'kd_tree', 'brute').
    metric : str
        Any distance metric accepted by :class:`sklearn.neighbors.NearestNeighbors`.
        If 'forest', then uses an :class:`treeple.UnsupervisedObliqueRandomForest`
        to compute geodesic distances.
    k : int, optional
        The number of k-nearest neighbors to query, by default 1.
    n_jobs : int,
        The number of CPUs to use for joblib. By default, None.

    Returns
    -------
    neigh : instance of sklearn.neighbors.NearestNeighbor
        A fitted instance of the nearest-neighbor algorithm on ``X`` input.

    Notes
    -----
    Can query for the following, using the ``neigh.kneighbors(X)`` function, which would
    return:

    dists : ArrayLike of shape (n_samples, k)
        The distance array of every sample with its k-nearest neighbors. The columns
        are ordered from closest to furthest neighbors.
    indices : ArrayLike of shape (n_samples, k)
        The sample indices of the k-nearest-neighbors for each sample. These
        contain the row indices of ``X`` for each sample. The columns
        are ordered from closest to furthest neighbors.
    """
    if metric == "forest":
        est = UnsupervisedObliqueRandomForest()
        dists = compute_forest_similarity_matrix(est, X)

        # we have a precomputed distance matrix, so we can use the NearestNeighbor
        # implementation of sklearn
        metric = "precomputed"
    else:
        dists = X

    # compute the nearest neighbors in the space using specified NN algorithm
    # then get the K nearest nbrs and their distances
    neigh = NearestNeighbors(n_neighbors=k, algorithm=algorithm, metric=metric, n_jobs=n_jobs).fit(
        dists
    )
    return neigh


def _trafo2uniform(X):
    """Transforms input array to uniform marginals.

    Assumes x.shape = (dim, T)

    Parameters
    ----------
    X : arraylike
        The input data with (n_samples,) rows and (n_features,) columns.

    Returns
    -------
    u : array-like
        array with uniform marginals.
    """

    def trafo(xi):
        xisorted = np.sort(xi)
        yi = np.linspace(1.0 / len(xi), 1, len(xi))
        return np.interp(xi, xisorted, yi)

    _, n_features = X.shape

    # apply a uniform transformation for each feature
    for idx in range(n_features):
        marginalized_feature = trafo(X[:, idx].to_numpy().squeeze())
        X[:, idx] = marginalized_feature
    return X



from .baseline import build_cv_forest, build_permutation_forest
from .forest import build_coleman_forest, build_oob_forest
from .permuteforest import PermutationHonestForestClassifier

__all__ = [
    "build_cv_forest",
    "build_oob_forest",
    "build_coleman_forest",
    "build_permutation_forest",
    "PermutationHonestForestClassifier",
]



"""
==============
Calculating MI
==============
"""

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from scipy.stats import entropy

from treeple.datasets import make_trunk_classification
from treeple.ensemble import HonestForestClassifier
from treeple.stats import build_oob_forest

sns.set(color_codes=True, style="white", context="talk", font_scale=1.5)
PALETTE = sns.color_palette("Set1")
sns.set_palette(PALETTE[1:5] + PALETTE[6:], n_colors=9)
sns.set_style("white", {"axes.edgecolor": "#dddddd"})
# %%
# MI
# --
#
# Mutual Information (*MI*) measures the mutual dependence between *X* and
# *Y*. It can be calculated by the difference between the class entropy
# (``H(Y)``) and the conditional entropy (``H(Y | X)``):
#
# .. math:: I(X; Y) = H(Y) - H(Y\mid X)
#
# With a binary class simulation as an example, this tutorial will show
# how to use ``treeple`` to calculate the statistic.

# %%
# Create a simulation with two gaussians
# --------------------------------------


# create a binary class simulation with two gaussians
# 500 samples for each class, class zero is standard
# gaussian, and class one has a mean at one
X, y = make_trunk_classification(
    n_samples=1000,
    n_dim=1,
    mu_0=0,
    mu_1=1,
    n_informative=1,
    seed=1,
)


fig, ax = plt.subplots(figsize=(6, 6))
fig.tight_layout()
ax.tick_params(labelsize=15)

# histogram plot the samples
ax.hist(X[:500], bins=50, alpha=0.6, color=PALETTE[1], label="negative")
ax.hist(X[500:], bins=50, alpha=0.3, color=PALETTE[0], label="positive")
ax.set_xlabel("Variable One", fontsize=15)
ax.set_ylabel("Likelihood", fontsize=15)
plt.legend(frameon=False, fontsize=15)
plt.show()


# %%
# Fit the model
# -------------


# initialize the forest with 100 trees
est = HonestForestClassifier(
    n_estimators=100,
    max_samples=1.6,
    max_features=0.3,
    bootstrap=True,
    stratify=True,
    random_state=1,
)

# fit the model and obtain the tree posteriors
_, observe_proba = build_oob_forest(est, X, y)

# generate forest posteriors for the two classes
observe_proba = np.nanmean(observe_proba, axis=0)


fig, ax = plt.subplots(figsize=(6, 6))
fig.tight_layout()
ax.tick_params(labelsize=15)

# histogram plot the posterior probabilities for class one
ax.hist(observe_proba[:500][:, 1], bins=50, alpha=0.6, color=PALETTE[1], label="negative")
ax.hist(observe_proba[500:][:, 1], bins=50, alpha=0.3, color=PALETTE[0], label="positive")
ax.set_ylabel("# of Samples", fontsize=15)
ax.set_xlabel("Class One Posterior", fontsize=15)
plt.legend(frameon=False, fontsize=15)
plt.show()


# %%
# Calculate the statistic
# -----------------------
def Calculate_MI(y_true, y_pred_proba):
    # calculate the conditional entropy
    H_YX = np.mean(entropy(y_pred_proba, base=np.exp(1), axis=1))

    # empirical count of each class (n_classes)
    _, counts = np.unique(y_true, return_counts=True)
    # calculate the entropy of labels
    H_Y = entropy(counts, base=np.exp(1))
    return H_Y - H_YX


mi = Calculate_MI(y, observe_proba)
print("MI =", round(mi, 2))



import pytest

from sklearn.base import BaseEstimator, RegressorMixin, TransformerMixin
from sklearn.utils._tags import get_tags


class NoTagsEstimator:
    pass


class ClassifierEstimator:
    # This is to test whether not inheriting from mixins works.
    _estimator_type = "classifier"


@pytest.mark.parametrize(
    "estimator, value",
    [
        [NoTagsEstimator(), False],
        [ClassifierEstimator(), True],
        [TransformerMixin(), False],
        [RegressorMixin(), True],
        [BaseEstimator(), False],
    ],
)
def test_requires_y(estimator, value):
    assert get_tags(estimator).target_tags.required == value



"""
========================================================
Predicting with different quantile interpolation methods
========================================================

An example comparison of interpolation methods that can be applied during
prediction when the desired quantile lies between two data points.

This example was heavily inspired by `quantile-forest <https://github.com/zillow/quantile-forest>`_
package. See their package `here <https://zillow.github.io/quantile-forest/>`_.
"""

from collections import defaultdict

import matplotlib.pyplot as plt
import numpy as np
from sklearn.ensemble import RandomForestRegressor

# %%
# Generate the data
# -----------------
# We use four simple data points to illustrate the difference between the intervals that are
# generated using different interpolation methods.

X = np.array([[-1, -1], [-1, -1], [-1, -1], [1, 1], [1, 1]])
y = np.array([-2, -1, 0, 1, 2])

# %%
# The interpolation methods
# -------------------------
# The following interpolation methods demonstrated here are:
# To interpolate between the data points, i and j (``i <= j``),
# linear, lower, higher, midpoint, or nearest. For more details, see `treeple.RandomForestRegressor`.
# The difference between the methods can be illustrated with the following example:

interpolations = ["linear", "lower", "higher", "midpoint", "nearest"]
colors = ["#006aff", "#ffd237", "#0d4599", "#f2a619", "#a6e5ff"]
quantiles = [0.025, 0.5, 0.975]

y_medians = []
y_errs = []
est = RandomForestRegressor(
    n_estimators=1,
    random_state=0,
)
# fit the model
est.fit(X, y)
# get the leaf nodes that each sample fell into
leaf_ids = est.apply(X)
# create a list of dictionary that maps node to samples that fell into it
# for each tree
node_to_indices = []
for tree in range(leaf_ids.shape[1]):
    d = defaultdict(list)
    for id, leaf in enumerate(leaf_ids[:, tree]):
        d[leaf].append(id)
    node_to_indices.append(d)
# drop the X_test to the trained tree and
# get the indices of leaf nodes that fall into it
leaf_ids_test = est.apply(X)
# for each samples, collect the indices of the samples that fell into
# the same leaf node for each tree
y_pred_quantile = []
for sample in range(leaf_ids_test.shape[0]):
    li = [
        node_to_indices[tree][leaf_ids_test[sample][tree]] for tree in range(leaf_ids_test.shape[1])
    ]
    # merge the list of indices into one
    idx = [item for sublist in li for item in sublist]
    # get the y_train for each corresponding id``
    y_pred_quantile.append(y[idx])

for interpolation in interpolations:
    # get the quatile preditions for each predicted sample
    y_pred = [
        np.array(
            [
                np.quantile(y_pred_quantile[i], quantile, method=interpolation)
                for i in range(len(y_pred_quantile))
            ]
        )
        for quantile in quantiles
    ]
    y_medians.append(y_pred[1])
    y_errs.append(
        np.concatenate(
            (
                [y_pred[1] - y_pred[0]],
                [y_pred[2] - y_pred[1]],
            ),
            axis=0,
        )
    )

sc = plt.scatter(np.arange(len(y)) - 0.35, y, color="k", zorder=10)
ebs = []
for i, (median, y_err) in enumerate(zip(y_medians, y_errs)):
    ebs.append(
        plt.errorbar(
            np.arange(len(y)) + (0.15 * (i + 1)) - 0.35,
            median,
            yerr=y_err,
            color=colors[i],
            ecolor=colors[i],
            fmt="o",
        )
    )
plt.xlim([-0.75, len(y) - 0.25])
plt.xticks(np.arange(len(y)), X.tolist())
plt.xlabel("Samples (Feature Values)")
plt.ylabel("Actual and Predicted Values")
plt.legend([sc] + ebs, ["actual"] + interpolations, loc=2)
plt.show()



"""
==========================================================================================
Plot oblique forest and axis-aligned random forest predictions on sparse parity simulation
==========================================================================================
A performance comparison between oblique forest and standard axis-
aligned random forest using sparse parity simulation dataset.
Sparse parity is a variation of the noisy parity problem,
which itself is a multivariate generalization of the noisy XOR problem.
This is a binary classification task in high dimensions. The simulation
will generate uniformly distributed `n_samples` number of sample points
in the range of -1 and +1 with `p` number of features. `p*` is a
parameter used to limit features that carry information about the class.
The informative binary label is then defined as 1 if there are odd number
of the sum of data `X` across first `p*` features that are greater than 0,
otherwise the label is defined as 0. The simulation is further detailed
in this [publication](https://epubs.siam.org/doi/epdf/10.1137/1.9781611974973.56).
"""

from datetime import datetime

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import RepeatedKFold, cross_validate

from treeple import ObliqueRandomForestClassifier

random_state = 123456
t0 = datetime.now()


def sparse_parity(n_samples, p=20, p_star=3, random_seed=None, **kwargs):
    if random_seed:
        np.random.seed(random_seed)

    X = np.random.uniform(-1, 1, (n_samples, p))
    y = np.zeros(n_samples)

    for i in range(0, n_samples):
        y[i] = sum(X[i, :p_star] > 0) % 2

    return X, y


def get_scores(X, y, n_cv=5, n_repeats=1, random_state=1, kwargs=None):
    clfs = [
        RandomForestClassifier(**kwargs[0], random_state=random_state),
        ObliqueRandomForestClassifier(**kwargs[1], random_state=random_state),
    ]

    tmp = []

    for i, clf in enumerate(clfs):
        cv = RepeatedKFold(n_splits=n_cv, n_repeats=n_repeats, random_state=random_state)
        test_score = cross_validate(estimator=clf, X=X, y=y, cv=cv, scoring="accuracy")

        tmp.append([["RF", "OF"][i], test_score["test_score"], test_score["test_score"].mean()])

    df = pd.DataFrame(tmp, columns=["model", "score", "mean"])
    df = df.explode("score")
    df["score"] = df["score"].astype(float)
    df.reset_index(inplace=True, drop=True)

    return df


# Grid searched hyper-parameters
params = [
    {"max_features": None, "n_estimators": 100, "max_depth": None},
    {"max_features": 40, "n_estimators": 100, "max_depth": 20},
]

X, y = sparse_parity(n_samples=1000, random_seed=random_state)

df = get_scores(X=X, y=y, n_cv=3, n_repeats=1, random_state=random_state, kwargs=params)
t_d = (datetime.now() - t0).seconds
print(f"It took {t_d} seconds to run the script")

# Draw a comparison plot
fig, ax = plt.subplots(1, 1, figsize=(6, 6))

sns.stripplot(data=df, x="model", y="score", ax=ax, dodge=True)
sns.boxplot(data=df, x="model", y="score", ax=ax, color="white")
ax.set_title("Sparse Parity")

rf = df.query('model=="RF"')["mean"].iloc[0]
rff = f"RF (Mean Test Score: {round(rf,3)})"

of = df.query('model=="OF"')["mean"].iloc[0]
off = f"OF (Mean Test Score: {round(of,3)})"

ax.legend([rff, off], loc=4)

plt.savefig(f"plot_sim_{t_d}s.jpg")
plt.show()
